//--------------------------------------------------------------------------
// Program to pull the information out of various types of EXIF digital 
// camera files and show it in a reasonably consistent way
//
// Version 2.3
//
//
// Compiling under Windows:  Use microsoft's compiler.  from command line:
// cl -Ox jhead.c exif.c myglob.c
//
// Dec 1999 - Jan 2005
//
// by Matthias Wandel   www.sentex.net/~mwandel
//--------------------------------------------------------------------------

// jheadwrap.c - based on jhead.c and exif.c
// Heiko Purnhagen 2005-02-04

#include "jhead.h"


int ShowTags     = FALSE;    // Do not show raw by default.


//--------------------------------------------------------------------------
// Error exit handler
//--------------------------------------------------------------------------
void ErrFatal(char * msg)
{
  fprintf(stderr,"Error : %s\n", msg);
  exit(EXIT_FAILURE);
} 

//--------------------------------------------------------------------------
// Report non fatal errors.  Now that microsoft.net modifies exif headers,
// there's corrupted ones, and there could be more in the future.
//--------------------------------------------------------------------------
void ErrNonfatal(char * msg, int a1, int a2)
{
#if 1
  return;
#else
  fprintf(stderr,"Nonfatal Error : ");
  fprintf(stderr, msg, a1, a2);
  fprintf(stderr, "\n");
#endif
} 


//--------------------------------------------------------------------------
// Summarize highlights of image info in one string
//--------------------------------------------------------------------------
char * GetConciseImageInfo(char * filename)
{
#define STRLEN 256
  static char s[STRLEN+1];
  char *c;
  ReadMode_t ReadMode;
  int i;

  ReadMode = READ_EXIF;

  ResetJpgfile();

  // Start with an empty image information structure.
  memset(&ImageInfo, 0, sizeof(ImageInfo));
  ImageInfo.FlashUsed = -1;
  ImageInfo.MeteringMode = -1;
  ImageInfo.Whitebalance = -1;

  i = ReadJpegFile(filename, ReadMode);
  if (!i) {
    snprintf(s,STRLEN,"EXIF error");
    return (s);
  }

  c = s;
  snprintf(c,STRLEN-(c-s),"%dx%d",ImageInfo.Width, ImageInfo.Height);
  while(*c)c++;

  if (ImageInfo.ExposureTime){
    if (ImageInfo.ExposureTime <= 0.105) {
      snprintf(c,STRLEN-(c-s),"  1/%d",(int)(0.5 + 1/ImageInfo.ExposureTime));
    } else if (ImageInfo.ExposureTime < 1) {
      snprintf(c,STRLEN-(c-s),"  1/%3.1f",1/ImageInfo.ExposureTime);
    } else if (ImageInfo.ExposureTime < 10) {
      snprintf(c,STRLEN-(c-s),"  %3.1f",ImageInfo.ExposureTime);
    } else {
      snprintf(c,STRLEN-(c-s),"  %d",(int)(0.5 + ImageInfo.ExposureTime));
    }
    while(*c)c++;
  }
  
  if (ImageInfo.ApertureFNumber){
    snprintf(c,STRLEN-(c-s),"  f/%3.1f",(double)ImageInfo.ApertureFNumber);
    while(*c)c++;
  }
  
  if (ImageInfo.FocalLength){
    snprintf(c,STRLEN-(c-s),"  %3.1fmm",ImageInfo.FocalLength);
    while(*c)c++;
  }
  
  if (ImageInfo.FocalLength35mmEquiv){
    snprintf(c,STRLEN-(c-s),"  (%dmm@35)",ImageInfo.FocalLength35mmEquiv);
    while(*c)c++;
  }
  
  if (ImageInfo.ExposureProgram){ // 05-jan-2001 vcs
    switch(ImageInfo.ExposureProgram) {
    case 1:
      snprintf(c,STRLEN-(c-s),"  M%+3.1f",(double)ImageInfo.ExposureBias);
      break;
    case 2:
      snprintf(c,STRLEN-(c-s),"  P%+3.1f",(double)ImageInfo.ExposureBias);
      break;
    case 3:
      snprintf(c,STRLEN-(c-s),"  A%+3.1f",(double)ImageInfo.ExposureBias);
      break;
    case 4:
      snprintf(c,STRLEN-(c-s),"  S%+3.1f",(double)ImageInfo.ExposureBias);
      break;
    default:
      break;
    }
    while(*c)c++;
  }
  
  if (ImageInfo.ISOequivalent){
    snprintf(c,STRLEN-(c-s),"  ISO:%2d",(int)ImageInfo.ISOequivalent);
    while(*c)c++;
  }
  
  if (ImageInfo.WhitebalanceName[0]){
    char *cc = ImageInfo.WhitebalanceName;
    while(*cc) { // strip trailing space
      if (*cc==' ')
	*cc = '\0';
      else
	cc++;
    }
    snprintf(c,STRLEN-(c-s),"  WB:%s", ImageInfo.WhitebalanceName);
    while(*c)c++;
  }
  
  if (ImageInfo.FlashUsed >= 0 && ImageInfo.FlashUsed & 1){
    snprintf(c,STRLEN-(c-s),"  (flash)");
    while(*c)c++;
  }
  
  if (ImageInfo.IsColor == 0){
    snprintf(c,STRLEN-(c-s),"  (bw)");
    while(*c)c++;
  }
  
  return (s);
}


#define NOMAIN
#ifndef NOMAIN
int main (int argc, char **argv)
{
  printf("returns: \"%s\"\n",GetConciseImageInfo(argv[1]));
  return 0;
}
#endif

