
/*
 * scaled idct
 * and YUV->RGB converter
 * 32 bit version
 *
 */

#include <stdio.h>
#include <string.h>

#define uchar unsigned char
#define uint64	unsigned long long


#define clip4(a,b,x)	a=b&0x01010101; b=0x01010101&(~(b>>7)); a*=255; b*=255; (x)|=a; (x)&=b

void sidct(uchar *y,short *x)
{
	int		z[64];
	register int	*zp;
	register int	n;
	register int	a0,a1,a2,a3,a4,a5,a6,a7;
	register int	b0,b1,b2,b3,b4,b5,b6,b7;

	n=8;
	zp=z;
	do
	{
		b0 = x[ 0];
		a1 = x[ 8];
		b2 = x[16];
		b3 = x[24];
		b1 = x[32];
		b5 = x[40];
		b6 = x[48];
	 	b7 = x[56];


		b4=a1+b7; a3=b3+b5; a2=b2-b6; a4=b5-b3;
		a5=b4-a3; a6=a1-b7; b3=b2+b6; a7=b4+a3;
		b7=a4+a6;

		a0=(196*b7)>>8;
		b2=(362*a2)>>8;
		b5=(362*a5)>>8;
		b4=(277*a4)>>8;
		b6=(669*a6)>>8;

		b4+=a0;
		b6-=a0;

		a2=b2-b3; a1=b0-b1; a3=b0+b1; a4=b6-a7;
		a0=a4-b5; a6=a1+a2; a5=a3+b3; b5=a1-a2;
		b6=a3-b3; b7=b4+a0;

		zp[0]=(a5+a7);
		zp[1]=(a6+a4);
		zp[2]=(b5-a0);
		zp[3]=(b6+b7);
		zp[4]=(b6-b7);
		zp[5]=(b5+a0);
		zp[6]=(a6-a4);
		zp[7]=(a5-a7);



		x++;
		zp+=8;
	}
	while (--n);

	x-=8;
	memset(x,0,64*sizeof(short));

	n=8;
	zp=z;

	do
	{
		b0 = zp[ 0];
		a1 = zp[ 8];
		b2 = zp[16];
		b3 = zp[24];
		b1 = zp[32];
		b5 = zp[40];
		b6 = zp[48];
	 	b7 = zp[56];

		b4=a1+b7; a3=b3+b5; a2=b2-b6; a4=b5-b3;
		a5=b4-a3; a6=a1-b7; b3=b2+b6; a7=b4+a3;
		b7=a4+a6;

		a0=(196*b7)>>8;
		b2=(362*a2)>>8;
		b5=(362*a5)>>8;
		b4=(277*a4)>>8;
		b6=(669*a6)>>8;

		b4+=a0;
		b6-=a0;

		a2=b2-b3; a1=b0-b1; a3=b0+b1; a4=b6-a7;
		a0=a4-b5; a6=a1+a2; a5=a3+b3; b5=a1-a2;
		b6=a3-b3; b7=b4+a0;

		a1=(a5+a7)>>6;
		a2=(a6+a4)>>6;
		a3=(b5-a0)>>6;
		b0=(b6+b7)>>6;
		b1=(b6-b7)>>6;
		b2=(b5+a0)>>6;
		b3=(a6-a4)>>6;
		b4=(a5-a7)>>6;

		if ( a1<0 ) a1=0;
		if ( a2<0 ) a2=0;
		if ( a3<0 ) a3=0;
		if ( b0<0 ) b0=0;
		if ( b1<0 ) b1=0;
		if ( b2<0 ) b2=0;
		if ( b3<0 ) b3=0;
		if ( b4<0 ) b4=0;

		if ( a1>255 ) a1=255;
		if ( a2>255 ) a2=255;
		if ( a3>255 ) a3=255;
		if ( b0>255 ) b0=255;
		if ( b1>255 ) b1=255;
		if ( b2>255 ) b2=255;
		if ( b3>255 ) b3=255;
		if ( b4>255 ) b4=255;

		y[0]=a1;
		y[1]=a2;
		y[2]=a3;
		y[3]=b0;
		y[4]=b1;
		y[5]=b2;
		y[6]=b3;
		y[7]=b4;

		zp++;
		y+=8;
	}
	while (--n);
}


void yuv2rgb_420(uchar *rgb,long d0,long d1,uchar *yuv)
{
	register uchar	*p;
	register long	i,j,y,u,v,h;

	for (i=0; i<16; i++)
	{
		p=rgb;
		for (j=0; j<16; j++)
		{
			y=yuv[8*(i&7)+(j&7)+16*(i&8)+8*(j&8)];
			u=yuv[256+8*(i>>1)+(j>>1)]-128;
			v=yuv[320+8*(i>>1)+(j>>1)]-128;
			h = y+((3629*u       +1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[0]=h;
			h = y-(( 705*u+1462*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[1]=h;
			h = y+((       2871*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[2]=h;
			p+=d0;
		}
		rgb+=d1;
	}
}

void yuv2rgb_422(uchar *rgb,long d0,long d1,uchar *yuv)
{
	register uchar	*p;
	register long	i,j,y,u,v,h;

	for (i=0; i<8; i++)
	{
		p=rgb;
		for (j=0; j<16; j++)
		{
			y=yuv[8*(i&7)+(j&7)+8*(j&8)];
			u=yuv[128+8*i+(j>>1)]-128;
			v=yuv[192+8*i+(j>>1)]-128;
			h = y+((3629*u       +1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[0]=h;
			h = y-(( 705*u+1462*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[1]=h;
			h = y+((       2871*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[2]=h;
			p+=d0;
		}
		rgb+=d1;
	}
}

void yuv2rgb_440(uchar *rgb,long d0,long d1,uchar *yuv)
{
	register uchar	*p;
	register long	i,j,y,u,v,h;

	for (i=0; i<16; i++)
	{
		p=rgb;
		for (j=0; j<8; j++)
		{
			y=yuv[8*i+j];
			u=yuv[128+8*(i>>1)+j]-128;
			v=yuv[192+8*(i>>1)+j]-128;
			h = y+((3629*u       +1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[0]=h;
			h = y-(( 705*u+1462*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[1]=h;
			h = y+((       2871*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[2]=h;
			p+=d0;
		}
		rgb+=d1;
	}
}

void yuv2rgb_444(uchar *rgb,long d0,long d1,uchar *yuv)
{
	register uchar	*p;
	register long	i,j,y,u,v,h;

	for (i=0; i<8; i++)
	{
		p=rgb;
		for (j=0; j<8; j++)
		{
			y=yuv[  0];
			u=yuv[ 64]-128;
			v=yuv[128]-128;

			h = y+((3629*u       +1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[0]=h;
			h = y-(( 705*u+1462*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[1]=h;
			h = y+((       2871*v+1024)>>11); if ( h<0 ) h=0; if ( h>255 ) h=255; p[2]=h;

			yuv++;
			p+=d0;
		}
		rgb+=d1;
	}
}
