
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sc.h"


#define clip(x)         if ( x<0 ) x=0; if ( x>255 ) x=255
#define sclip(x,s)      x=(x+(1<<(s-1)))>>s; if ( x<0 ) x=0; if ( x>255 ) x=255

static void scale_1_4_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 : y[0]=x[0]<<6;
				 y[1]=x[1]<<6;
				 y[2]=x[2]<<6;
				 px=1;
				 y+=3;
				 continue;

			case 1 : y[0]=-6*x[-4]+56*x[0]+16*x[4]-2*x[8];
				 y[1]=-6*x[-3]+56*x[1]+16*x[5]-2*x[9];
				 y[2]=-6*x[-2]+56*x[2]+16*x[6]-2*x[10];
				 px=2;
				 y+=3;
				 continue;

			case 2 : y[0]=37*(x[0]+x[4])-5*(x[-4]+x[8]);
				 y[1]=37*(x[1]+x[5])-5*(x[-3]+x[9]);
				 y[2]=37*(x[2]+x[6])-5*(x[-2]+x[10]);
				 px=3;
				 y+=3;
				 continue;

			case 3 : y[0]=-2*x[-4]+16*x[0]+56*x[4]-6*x[8];
				 y[1]=-2*x[-3]+16*x[1]+56*x[5]-6*x[9];
				 y[2]=-2*x[-2]+16*x[2]+56*x[6]-6*x[10];
				 px=0;
				 x+=4;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_1_4_ver_0(uchar *y,int *x,long n)
{
	register long	a,b,c;

	do
	{
		a=(x[0]+32)>>6; clip(a);
		b=(x[1]+32)>>6; clip(b);
		c=(x[2]+32)>>6; clip(c);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static void scale_1_4_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=(-3*x0[0]+28*x1[0]+8*x2[0]-x3[0]); sclip(a,11);
		b=(-3*x0[1]+28*x1[1]+8*x2[1]-x3[1]); sclip(b,11);
		c=(-3*x0[2]+28*x1[2]+8*x2[2]-x3[2]); sclip(c,11);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_1_4_ver_2(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=37*(x1[0]+x2[0])-5*(x0[0]+x3[0]); sclip(a,12);
		b=37*(x1[1]+x2[1])-5*(x0[1]+x3[1]); sclip(b,12);
		c=37*(x1[2]+x2[2])-5*(x0[2]+x3[2]); sclip(c,12);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_1_4(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[4];
	register long	i,j;

	dx*=4;
	x+=4*((px>>2)+border)+((py>>2)+border-1)*dx;
	px&=3;
	py&=3;

	for (i=0; i<4; i++)
	{
		buf[i]=tb;
		scale_1_4_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}

	for (j=py; j<py+h; j++)
	{
		switch ( j&3 ) {
			case 0 : scale_1_4_ver_0(y,       buf[1],              w); y+=dy;	continue;
			case 1 : scale_1_4_ver_1(y,buf[0],buf[1],buf[2],buf[3],w); y+=dy;	continue;
			case 2 : scale_1_4_ver_2(y,buf[0],buf[1],buf[2],buf[3],w); y+=dy;	continue;
			case 3 : scale_1_4_ver_1(y,buf[3],buf[2],buf[1],buf[0],w); y+=dy;
				 tb=buf[0]; for (i=0; i<3; i++) buf[i]=buf[i+1]; buf[3]=tb;
				 scale_1_4_hor(tb,x,px,w); x+=dx;
				 continue;
		}
	}
}

static void scale_1_3_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 : y[0]=x[0]<<7;
				 y[1]=x[1]<<7;
				 y[2]=x[2]<<7;
				 px=1;
				 y+=3;
				 continue;

			case 1 : y[0]=5*x[-8]-20*x[-4]+104*x[0]+49*x[4]-13*x[8]+3*x[12];
				 y[1]=5*x[-7]-20*x[-3]+104*x[1]+49*x[5]-13*x[9]+3*x[13];
				 y[2]=5*x[-6]-20*x[-2]+104*x[2]+49*x[6]-13*x[10]+3*x[14];
				 px=2;
				 y+=3;
				 continue;

			case 2 : y[0]=3*x[-8]-13*x[-4]+49*x[0]+104*x[4]-20*x[8]+5*x[12];
				 y[1]=3*x[-7]-13*x[-3]+49*x[1]+104*x[5]-20*x[9]+5*x[13];
				 y[2]=3*x[-6]-13*x[-2]+49*x[2]+104*x[6]-20*x[10]+5*x[14];
				 px=0;
				 x+=4;
				 y+=3;
				 continue;
		}
	}
}

static void scale_1_3_ver_0(uchar *y,int *x,long n)
{
	register long	a,b,c;

	do
	{
		a=(x[0]+64)>>7; clip(a);
		b=(x[1]+64)>>7; clip(b);
		c=(x[2]+64)>>7; clip(c);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static void scale_1_3_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,long n)
{
	register long	a,b,c;

	do
	{
		a=5*x0[0]-20*x1[0]+104*x2[0]+49*x3[0]-13*x4[0]+3*x5[0]; sclip(a,14);
		b=5*x0[1]-20*x1[1]+104*x2[1]+49*x3[1]-13*x4[1]+3*x5[1]; sclip(b,14);
		c=5*x0[2]-20*x1[2]+104*x2[2]+49*x3[2]-13*x4[2]+3*x5[2]; sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3;
		x1+=3;
		x2+=3;
		x3+=3;
		x4+=3;
		x5+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_1_3(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[6];
	register long	i;

	dx*=4;
	x+=4*((px/3)+border)+((py/3)+border-2)*dx;
	px%=3;
	py%=3;

	for (i=0; i<6; i++)
	{
		buf[i]=tb;
		scale_1_3_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}

	for (; h; h--)
	{
		switch ( py ) {
			case 0 : scale_1_3_ver_0(y,              buf[2],                     w); y+=dy; py=1;	continue;
			case 1 : scale_1_3_ver_1(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],w); y+=dy; py=2;	continue;
			case 2 : scale_1_3_ver_1(y,buf[5],buf[4],buf[3],buf[2],buf[1],buf[0],w); y+=dy; py=0;
				 tb=buf[0]; for (i=0; i<5; i++) buf[i]=buf[i+1]; buf[5]=tb;
				 scale_1_3_hor(tb,x,px,w); x+=dx;
				 continue;
		}
	}
}

static void scale_1_2_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 :
				 y[0]=x[0]<<6;
				 y[1]=x[1]<<6;
				 y[2]=x[2]<<6;
				 px=1;
				 y+=3;
				 continue;
			case 1 :
				 y[0]=39*(x[0]+x[4])-9*(x[-4]+x[8])+2*(x[-8]+x[12]);
				 y[1]=39*(x[1]+x[5])-9*(x[-3]+x[9])+2*(x[-7]+x[13]);
				 y[2]=39*(x[2]+x[6])-9*(x[-2]+x[10])+2*(x[-6]+x[14]);
				 px=0;
				 x+=4;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_1_2_ver_0(uchar *y,int *x,long n)
{
	register long	a,b,c;

	do
	{
		a=(x[0]+32)>>6; clip(a);
		b=(x[1]+32)>>6; clip(b);
		c=(x[2]+32)>>6; clip(c);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_1_2_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,long n)
{
	register long	a,b,c;

	do
	{
		a=39*(x2[0]+x3[0])-9*(x1[0]+x4[0])+2*(x0[0]+x5[0]); sclip(a,12);
		b=39*(x2[1]+x3[1])-9*(x1[1]+x4[1])+2*(x0[1]+x5[1]); sclip(b,12);
		c=39*(x2[2]+x3[2])-9*(x1[2]+x4[2])+2*(x0[2]+x5[2]); sclip(c,12);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_1_2(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[6];
	register long	i;

	dx*=4;
	x+=4*((px>>1)+border)+((py>>1)+border-2)*dx;
	px&=1;
	py&=1;

	for (i=0; i<6; i++)
	{
		buf[i]=tb;
		scale_1_2_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}

	for (; h; h--)
	{
		switch ( py ) {
			case 0 : scale_1_2_ver_0(y,              buf[2],                     w); y+=dy; py=1;	continue;
			case 1 : scale_1_2_ver_1(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],w); y+=dy; py=0;
				 tb=buf[0]; for (i=0; i<5; i++) buf[i]=buf[i+1]; buf[5]=tb;
				 scale_1_2_hor(tb,x,px,w); x+=dx;
				 continue;
		}
	}
}


static void scale_2_3_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 : y[0]=x[0]<<7;
				 y[1]=x[1]<<7;
				 y[2]=x[2]<<7;
				 px=2;
				 y+=3;
				 continue;

			case 1 : y[0]=5*x[-8]-20*x[-4]+104*x[0]+49*x[4]-13*x[8]+3*x[12];
				 y[1]=5*x[-7]-20*x[-3]+104*x[1]+49*x[5]-13*x[9]+3*x[13];
				 y[2]=5*x[-6]-20*x[-2]+104*x[2]+49*x[6]-13*x[10]+3*x[14];
				 px=0;
				 x+=4;
				 y+=3;
				 continue;

			case 2 : y[0]=3*x[-8]-13*x[-4]+49*x[0]+104*x[4]-20*x[8]+5*x[12];
				 y[1]=3*x[-7]-13*x[-3]+49*x[1]+104*x[5]-20*x[9]+5*x[13];
				 y[2]=3*x[-6]-13*x[-2]+49*x[2]+104*x[6]-20*x[10]+5*x[14];
				 px=1;
				 x+=4;
				 y+=3;
				 continue;
		}
	}
}

void scale_2_3(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[6];
	register long	i;

	dx*=4;
	px*=2;
	py*=2;
	x+=4*(px/3+border)+(py/3+border-2)*dx;
	px%=3;
	py%=3;

	for (i=0; i<6; i++)
	{
		buf[i]=tb;
		scale_2_3_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}

	for (; h; h--)
	{
		switch ( py ) {
			case 0 : scale_1_3_ver_0(y,              buf[2],                     w); y+=dy; py=2;
				 continue;
			case 1 : scale_1_3_ver_1(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],w); y+=dy; py=0;
				 tb=buf[0]; for (i=0; i<5; i++) buf[i]=buf[i+1]; buf[5]=tb;
				 scale_2_3_hor(tb,x,px,w); x+=dx;
				 continue;
			case 2 : scale_1_3_ver_1(y,buf[5],buf[4],buf[3],buf[2],buf[1],buf[0],w); y+=dy; py=1;
				 tb=buf[0]; for (i=0; i<5; i++) buf[i]=buf[i+1]; buf[5]=tb;
				 scale_2_3_hor(tb,x,px,w); x+=dx;
				 continue;
		}
	}
}


void scale_1_1(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	dx*=4;
	x+=4*(px+border)+(py+border)*dx;

	do
	{
		memcpy(y,x,4*w);
		x+=dx;
		y+=dy;
		h--;
	}
	while ( h );	
}



static void scale_6_5_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 :
				 y[0]=106*x[0]+16*(x[-4]+x[4])-5*(x[-8]+x[8]);
				 y[1]=106*x[1]+16*(x[-3]+x[5])-5*(x[-7]+x[9]);
				 y[2]=106*x[2]+16*(x[-2]+x[6])-5*(x[-6]+x[10]);
				 px=1;
				 x+=4;
				 y+=3;
				 continue;
			case 1 :
				 y[0]=-2*x[-8]+101*x[0]+38*x[4]-9*x[8];
				 y[1]=-2*x[-7]+101*x[1]+38*x[5]-9*x[9];
				 y[2]=-2*x[-6]+101*x[2]+38*x[6]-9*x[10];
				 px=2;
				 x+=4;
				 y+=3;
				 continue;
			case 2 :
				 y[0]=-8*x[-4]+84*x[0]+62*x[4]-10*x[8];
				 y[1]=-8*x[-3]+84*x[1]+62*x[5]-10*x[9];
				 y[2]=-8*x[-2]+84*x[2]+62*x[6]-10*x[10];
				 px=3;
				 x+=4;
				 y+=3;
				 continue;
			case 3 :
				 y[0]=-10*x[-4]+62*x[0]+84*x[4]-8*x[8];
				 y[1]=-10*x[-3]+62*x[1]+84*x[5]-8*x[9];
				 y[2]=-10*x[-2]+62*x[2]+84*x[6]-8*x[10];
				 px=4;
				 x+=4;
				 y+=3;
				 continue;
			case 4 :
				 y[0]=-9*x[-4]+38*x[0]+101*x[4]-2*x[12];
				 y[1]=-9*x[-3]+38*x[1]+101*x[5]-2*x[13];
				 y[2]=-9*x[-2]+38*x[2]+101*x[6]-2*x[14];
				 px=0;
				 x+=8;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_6_5_ver_0(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,long n)
{
	register long	a,b,c;

	do
	{
		a=106*x2[0]+16*(x1[0]+x3[0])-5*(x0[0]+x4[0]); sclip(a,14);
		b=106*x2[1]+16*(x1[1]+x3[1])-5*(x0[1]+x4[1]); sclip(b,14);
		c=106*x2[2]+16*(x1[2]+x3[2])-5*(x0[2]+x4[2]); sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_6_5_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=-2*x0[0]+101*x1[0]+38*x2[0]-9*x3[0]; sclip(a,14);
		b=-2*x0[1]+101*x1[1]+38*x2[1]-9*x3[1]; sclip(b,14);
		c=-2*x0[2]+101*x1[2]+38*x2[2]-9*x3[2]; sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_6_5_ver_2(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=-8*x0[0]+84*x1[0]+62*x2[0]-10*x3[0]; sclip(a,14);
		b=-8*x0[1]+84*x1[1]+62*x2[1]-10*x3[1]; sclip(b,14);
		c=-8*x0[2]+84*x1[2]+62*x2[2]-10*x3[2]; sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_6_5(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[6];
	register long	i;

	dx*=4;
	px*=6;
	py*=6;
	x+=4*((px/5)+border)+((py/5)+border-2)*dx;
	px%=5;
	py%=5;

	for (i=0; i<5; i++)
	{
		buf[i]=tb;
		scale_6_5_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}
	buf[5]=tb;

	for (; h; h--)
	{
		scale_6_5_hor(tb,x,px,w); x+=dx;

		switch ( py ) {
			case 0 : scale_6_5_ver_0(y,buf[0],buf[1],buf[2],buf[3],buf[4],w);	y+=dy; py=1;  break;
			case 1 : scale_6_5_ver_1(y,buf[0],buf[2],buf[3],buf[4],w);		y+=dy; py=2;  break;
			case 2 : scale_6_5_ver_2(y,buf[1],buf[2],buf[3],buf[4],w);		y+=dy; py=3;  break;
			case 3 : scale_6_5_ver_2(y,buf[4],buf[3],buf[2],buf[1],w);		y+=dy; py=4;  break;
			case 4 : scale_6_5_ver_1(y,buf[5],buf[3],buf[2],buf[1],w);		y+=dy; py=0;  break;
		}

		tb=buf[0]; for (i=0; i<5; i++) buf[i]=buf[i+1]; buf[5]=tb;

		if ( !py )
		{
			scale_6_5_hor(tb,x,px,w); x+=dx;
			tb=buf[0]; for (i=0; i<5; i++) buf[i]=buf[i+1]; buf[5]=tb;
		}
	}
}


static void scale_4_3_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 :
				 y[0]=96*x[0]+25*(x[-4]+x[4])-11*(x[-8]+x[8])+2*(x[-12]+x[12]);
				 y[1]=96*x[1]+25*(x[-3]+x[5])-11*(x[-7]+x[9])+2*(x[-11]+x[13]);
				 y[2]=96*x[2]+25*(x[-2]+x[6])-11*(x[-6]+x[10])+2*(x[-10]+x[14]);
				 px=1;
				 x+=4;
				 y+=3;
				 continue;

			case 1 :
				 y[0]=x[-12]-5*x[-8]+85*x[0]+58*x[4]-11*x[8];
				 y[1]=x[-11]-5*x[-7]+85*x[1]+58*x[5]-11*x[9];
				 y[2]=x[-10]-5*x[-6]+85*x[2]+58*x[6]-11*x[10];
				 px=2;
				 x+=4;
				 y+=3;
				 continue;

			case 2 :
				 y[0]=-11*x[-4]+58*x[0]+85*x[4]-5*x[12]+x[16];
				 y[1]=-11*x[-3]+58*x[1]+85*x[5]-5*x[13]+x[17];
				 y[2]=-11*x[-2]+58*x[2]+85*x[6]-5*x[14]+x[18];

				 px=0;
				 x+=8;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_4_3_ver_0(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,long n)
{
	register long	a,b,c;

	do
	{
		a=96*x3[0]+25*(x2[0]+x4[0])-11*(x1[0]+x5[0])+2*(x0[0]+x6[0]); sclip(a,14);
		b=96*x3[1]+25*(x2[1]+x4[1])-11*(x1[1]+x5[1])+2*(x0[1]+x6[1]); sclip(b,14);
		c=96*x3[2]+25*(x2[2]+x4[2])-11*(x1[2]+x5[2])+2*(x0[2]+x6[2]); sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_4_3_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,long n)
{
	register long	a,b,c;

	do
	{
		a=x0[0]-5*x1[0]+85*x2[0]+58*x3[0]-11*x4[0]; sclip(a,14);
		b=x0[1]-5*x1[1]+85*x2[1]+58*x3[1]-11*x4[1]; sclip(b,14);
		c=x0[2]-5*x1[2]+85*x2[2]+58*x3[2]-11*x4[2]; sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_4_3(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[8];
	register long	i;

	dx*=4;
	px*=4;
	py*=4;
	x+=4*((px/3)+border)+((py/3)+border-3)*dx;
	px%=3;
	py%=3;

	for (i=0; i<7; i++)
	{
		buf[i]=tb;
		scale_4_3_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}
	buf[7]=tb;

	for (; h; h--)
	{
		scale_4_3_hor(tb,x,px,w); x+=dx;
		
		switch ( py ) {
			case 0 : scale_4_3_ver_0(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],w); y+=dy; py=1;  break;
			case 1 : scale_4_3_ver_1(y,buf[0],buf[1],       buf[3],buf[4],buf[5],       w); y+=dy; py=2;  break;
			case 2 : scale_4_3_ver_1(y,buf[7],buf[6],       buf[4],buf[3],buf[2],       w); y+=dy; py=0;  break;
		}

		tb=buf[0]; for (i=0; i<7; i++) buf[i]=buf[i+1]; buf[7]=tb;

		if ( !py )
		{
			scale_4_3_hor(tb,x,px,w); x+=dx;
			tb=buf[0]; for (i=0; i<7; i++) buf[i]=buf[i+1]; buf[7]=tb;
		}
	}
}



static void scale_3_2_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 : y[0]=86*x[0]+31*(x[-4]+x[4])-10*(x[-8]+x[8]);
				 y[1]=86*x[1]+31*(x[-3]+x[5])-10*(x[-7]+x[9]);
				 y[2]=86*x[2]+31*(x[-2]+x[6])-10*(x[-6]+x[10]);
				 px=1;
				 x+=4;
				 y+=3;
				 continue;

			case 1 : y[0]=(17*(x[0]+x[4])-x[-8]-x[12])<<2;
				 y[1]=(17*(x[1]+x[5])-x[-7]-x[13])<<2;
				 y[2]=(17*(x[2]+x[6])-x[-6]-x[14])<<2;
				 px=0;
				 x+=8;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_3_2_ver_0(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,long n)
{
	register long	a,b,c;

	do
	{
		a=86*x2[0]+31*(x1[0]+x3[0])-10*(x0[0]+x4[0]); sclip(a,14);
		b=86*x2[1]+31*(x1[1]+x3[1])-10*(x0[1]+x4[1]); sclip(b,14);
		c=86*x2[2]+31*(x1[2]+x3[2])-10*(x0[2]+x4[2]); sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_3_2_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=17*(x1[0]+x2[0])-x0[0]-x3[0]; sclip(a,12);
		b=17*(x1[1]+x2[1])-x0[1]-x3[1]; sclip(b,12);
		c=17*(x1[2]+x2[2])-x0[2]-x3[2]; sclip(c,12);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_3_2(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[5];
	register long	i;

	dx*=4;
	px*=3;
	py*=3;
	x+=4*((px>>1)+border)+((py>>1)+border-2)*dx;
	px&=1;
	py&=1;

	for (i=0; i<4; i++)
	{
		buf[i]=tb;
		scale_3_2_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}
	buf[4]=tb;

	for (; h; h--)
	{
		scale_3_2_hor(tb,x,px,w); x+=dx;
		
		switch ( py ) {
			case 0 : scale_3_2_ver_0(y,buf[0],buf[1],buf[2],buf[3],buf[4],w); y+=dy; py=1;  break;
			case 1 : scale_3_2_ver_1(y,buf[1],buf[2],buf[3],buf[4]       ,w); y+=dy; py=0;  break;
		}

		tb=buf[0]; for (i=0; i<4; i++) buf[i]=buf[i+1]; buf[4]=tb;

		if ( !py )
		{
			scale_3_2_hor(tb,x,px,w); x+=dx;
			tb=buf[0]; for (i=0; i<4; i++) buf[i]=buf[i+1]; buf[4]=tb;
		}
	}
}


static void scale_5_3_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 :
				 y[0]=78*x[0]+34*(x[-4]+x[4])-7*(x[-8]+x[8])-2*(x[-12]+x[12]);
				 y[1]=78*x[1]+34*(x[-3]+x[5])-7*(x[-7]+x[9])-2*(x[-11]+x[13]);
				 y[2]=78*x[2]+34*(x[-2]+x[6])-7*(x[-6]+x[10])-2*(x[-10]+x[14]);
				 px=2;
				 x+=4;
				 y+=3;
				 continue;
			case 1 :
				 y[0]=-7*x[-8]+14*x[-4]+71*x[0]+55*x[4]-5*x[12];
				 y[1]=-7*x[-7]+14*x[-3]+71*x[1]+55*x[5]-5*x[13];
				 y[2]=-7*x[-6]+14*x[-2]+71*x[2]+55*x[6]-5*x[14];
				 px=0;
				 x+=8;
				 y+=3;
				 continue;

			case 2 :
				 y[0]=-5*x[-8]+55*x[0]+71*x[4]+14*x[8]-7*x[12];
				 y[1]=-5*x[-7]+55*x[1]+71*x[5]+14*x[9]-7*x[13];
				 y[2]=-5*x[-6]+55*x[2]+71*x[6]+14*x[10]-7*x[14];
				 px=1;
				 x+=8;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_5_3_ver_0(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,long n)
{
	register long	a,b,c;

	do
	{
		a=78*x3[0]+34*(x2[0]+x4[0])-7*(x1[0]+x5[0])-2*(x0[0]+x6[0]); sclip(a,14);
		b=78*x3[1]+34*(x2[1]+x4[1])-7*(x1[1]+x5[1])-2*(x0[1]+x6[1]); sclip(b,14);
		c=78*x3[2]+34*(x2[2]+x4[2])-7*(x1[2]+x5[2])-2*(x0[2]+x6[2]); sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_5_3_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,long n)
{
	register long	a,b,c;

	do
	{
		a=-7*x0[0]+14*x1[0]+71*x2[0]+55*x3[0]-5*x4[0]; sclip(a,14);
		b=-7*x0[1]+14*x1[1]+71*x2[1]+55*x3[1]-5*x4[1]; sclip(b,14);
		c=-7*x0[2]+14*x1[2]+71*x2[2]+55*x3[2]-5*x4[2]; sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_5_3(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[7];
	register long	i;

	dx*=4;
	px*=5;
	py*=5;
	x+=4*((px/3)+border)+((py/3)+border-3)*dx;
	px%=3;
	py%=3;

	for (i=0; i<6; i++)
	{
		buf[i]=tb;
		scale_5_3_hor(tb,x,px,w); x+=dx;
		tb+=3*w;
	}
	buf[6]=tb;

	for (; h; h--)
	{
		scale_5_3_hor(tb,x,px,w); x+=dx;
		
		switch ( py ) {
			case 0 : scale_5_3_ver_0(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],w); y+=dy; py=2;  break;
			case 1 : scale_5_3_ver_1(y,buf[1],buf[2],buf[3],buf[4],buf[6],w);		y+=dy; py=0;  break;
			case 2 : scale_5_3_ver_1(y,buf[6],buf[5],buf[4],buf[3],buf[1],w);		y+=dy; py=1;  break;
		}

		tb=buf[0]; for (i=0; i<6; i++) buf[i]=buf[i+1]; buf[6]=tb;

		if ( py!=2 )
		{
			scale_5_3_hor(tb,x,px,w); x+=dx;
			tb=buf[0]; for (i=0; i<6; i++) buf[i]=buf[i+1]; buf[6]=tb;
		}
	}
}

static void scale_2_1_hor(int *y,uchar *x,long n)
{
	register long	a,b,c;

	do
	{
		a=x[-4]+x[4]; y[0]=32*x[0]+16*a+3*(a-x[-12]-x[12]);
		b=x[-3]+x[5]; y[1]=32*x[1]+16*b+3*(b-x[-11]-x[13]);
		c=x[-2]+x[6]; y[2]=32*x[2]+16*c+3*(c-x[-10]-x[14]);
		x+=8;
		y+=6;
		n--;
	}
	while ( n );
}

static inline void scale_2_1_ver(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b;
	do
	{
		b=x1[0]+x2[0]; a=32*x1[3]+16*b+3*(b-x0[0]-x3[0]); sclip(a,12); y[0]=a;
		a=x1[1]+x2[1]; b=32*x1[4]+16*a+3*(a-x0[1]-x3[1]); sclip(b,12); y[1]=b;
		b=x1[2]+x2[2]; a=32*x1[5]+16*b+3*(b-x0[2]-x3[2]); sclip(a,12); y[2]=a;
		y[3]=0;
		x0+=6; x1+=6; x2+=6; x3+=6;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_2_1(int *tb0,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	register int	*tb1,*tb2,*tb3, *ttb;
	register long	i;

	dx*=4;
	x+=4*(2*px+border)+(2*py+border-3)*dx;

	tb1=tb0+6*w;	for (i=0; i<2; i++) { scale_2_1_hor(tb0+3*i,x,w); x+=dx; }
	tb2=tb1+6*w;	for (i=0; i<2; i++) { scale_2_1_hor(tb1+3*i,x,w); x+=dx; }
	tb3=tb2+6*w;	for (i=0; i<2; i++) { scale_2_1_hor(tb2+3*i,x,w); x+=dx; }

	do
	{
		scale_2_1_hor(tb3  ,x,w); x+=dx;
		scale_2_1_hor(tb3+3,x,w); x+=dx;
		scale_2_1_ver(y,tb0,tb1,tb2,tb3,w); y+=dy;
		ttb=tb0;
		tb0=tb1; tb1=tb2; tb2=tb3; tb3=ttb;
		h--;
	}
	while ( h );
}

static void scale_9_4_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 :
				 y[0]=114*x[0]+73*(x[-4]+x[4])+9*(x[-8]+x[8])-9*(x[-12]+x[12])-2*(x[-16]+x[16]);
				 y[1]=114*x[1]+73*(x[-3]+x[5])+9*(x[-7]+x[9])-9*(x[-11]+x[13])-2*(x[-15]+x[17]);
				 y[2]=114*x[2]+73*(x[-2]+x[6])+9*(x[-6]+x[10])-9*(x[-10]+x[14])-2*(x[-14]+x[18]);
				 px=1;
				 x+=8;
				 y+=3;
				 continue;
			case 1 :
				 y[0]=-8*x[-12]+56*x[-4]+110*x[0]+89*x[4]+22*x[8]-9*x[12]-4*x[16];
				 y[1]=-8*x[-11]+56*x[-3]+110*x[1]+89*x[5]+22*x[9]-9*x[13]-4*x[17];
				 y[2]=-8*x[-10]+56*x[-2]+110*x[2]+89*x[6]+22*x[10]-9*x[14]-4*x[18];
				 px=2;
				 x+=8;
				 y+=3;
				 continue;

			case 2 :
				 y[0]=102*(x[0]+x[4])+38*(x[-4]+x[8])-6*(x[-12]+x[-8]+x[12]+x[16]);
				 y[1]=102*(x[1]+x[5])+38*(x[-3]+x[9])-6*(x[-11]+x[-7]+x[13]+x[17]);
				 y[2]=102*(x[2]+x[6])+38*(x[-2]+x[10])-6*(x[-10]+x[-6]+x[14]+x[18]);
				 px=3;
				 x+=8;
				 y+=3;
				 continue;

			case 3 :
				 y[0]=-4*x[-12]-9*x[-8]+22*x[-4]+89*x[0]+110*x[4]+56*x[8]-8*x[16];
				 y[1]=-4*x[-11]-9*x[-7]+22*x[-3]+89*x[1]+110*x[5]+56*x[9]-8*x[17];
				 y[2]=-4*x[-10]-9*x[-6]+22*x[-2]+89*x[2]+110*x[6]+56*x[10]-8*x[18];
				 px=0;
				 x+=12;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_9_4_ver_0(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,int *x7,int *x8,long n)
{
	register long	a,b,c;

	do
	{
		a=114*x4[0]+73*(x3[0]+x5[0])+9*(x2[0]+x6[0])-9*(x1[0]+x7[0])-2*(x0[0]+x8[0]); sclip(a,16);
		b=114*x4[1]+73*(x3[1]+x5[1])+9*(x2[1]+x6[1])-9*(x1[1]+x7[1])-2*(x0[1]+x8[1]); sclip(b,16);
		c=114*x4[2]+73*(x3[2]+x5[2])+9*(x2[2]+x6[2])-9*(x1[2]+x7[2])-2*(x0[2]+x8[2]); sclip(c,16);

		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3; x7+=3; x8+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_9_4_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,long n)
{
	register long	a,b,c;

	do
	{
		a=-8*x0[0]+56*x1[0]+110*x2[0]+89*x3[0]+22*x4[0]-9*x5[0]-4*x6[0]; sclip(a,16);
		b=-8*x0[1]+56*x1[1]+110*x2[1]+89*x3[1]+22*x4[1]-9*x5[1]-4*x6[1]; sclip(b,16);
		c=-8*x0[2]+56*x1[2]+110*x2[2]+89*x3[2]+22*x4[2]-9*x5[2]-4*x6[2]; sclip(c,16);

		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_9_4_ver_2(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,int *x7,long n)
{
	register long	a,b,c;

	do
	{
		a=102*(x3[0]+x4[0])+38*(x2[0]+x5[0])-6*(x0[0]+x1[0]+x6[0]+x7[0]); sclip(a,16);
		b=102*(x3[1]+x4[1])+38*(x2[1]+x5[1])-6*(x0[1]+x1[1]+x6[1]+x7[1]); sclip(b,16);
		c=102*(x3[2]+x4[2])+38*(x2[2]+x5[2])-6*(x0[2]+x1[2]+x6[2]+x7[2]); sclip(c,16);

		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3; x7+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_9_4(int *tb0,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[9];
	register int	*tb1,*tbh;
	register long	i;

	dx*=4;
	px*=9;
	py*=9;
	x+=4*((px>>2)+border)+((py>>2)+border-4)*dx;
	px&=3;
	py&=3;

	for (i=0; i<7; i++)
	{
		buf[i]=tb0;
		scale_9_4_hor(tb0,x,px,w); x+=dx;
		tb0+=3*w;
	}
	buf[7]=tb0;	tb1=tb0+3*w;
	buf[8]=tb1;

	for (; h; h--)
	{
		scale_9_4_hor(tb0,x,px,w); x+=dx;
		scale_9_4_hor(tb1,x,px,w); x+=dx;
		
		switch ( py ) {
			case 0 : scale_9_4_ver_0(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],buf[7],buf[8],w);	y+=dy; py=1;	break;
			case 1 : scale_9_4_ver_1(y,buf[1],buf[3],buf[4],buf[5],buf[6],buf[7],buf[8],w);			y+=dy; py=2;	break;
			case 2 : scale_9_4_ver_2(y,buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],buf[7],buf[8],w);		y+=dy; py=3;	break;
			case 3 : scale_9_4_ver_1(y,buf[8],buf[6],buf[5],buf[4],buf[3],buf[2],buf[1],w);			y+=dy; py=0;	break;
		}

		if ( py )
		{
			tb0=buf[0];
			tb1=buf[1];
			for (i=0; i<7; i++) buf[i]=buf[i+2];
			buf[7]=tb0;
			buf[8]=tb1;
		}
		else
		{
			tbh=buf[0];
			tb0=buf[1];
			tb1=buf[2];
			for (i=0; i<6; i++) buf[i]=buf[i+3];
			buf[6]=tbh;
			buf[7]=tb0;
			buf[8]=tb1;
			scale_9_4_hor(tbh,x,px,w); x+=dx;
		}
	}
}



static void scale_5_2_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 :
				 y[0]=52*x[0]+36*(x[-4]+x[4])+9*(x[-8]+x[8])-4*(x[-12]+x[12])-3*(x[-16]+x[16]);
				 y[1]=52*x[1]+36*(x[-3]+x[5])+9*(x[-7]+x[9])-4*(x[-11]+x[13])-3*(x[-15]+x[17]);
				 y[2]=52*x[2]+36*(x[-2]+x[6])+9*(x[-6]+x[10])-4*(x[-10]+x[14])-3*(x[-14]+x[18]);
				 px=1;
				 x+=8;
				 y+=3;
				 continue;
			case 1 :
				 y[0]=47*(x[0]+x[4])+22*(x[-4]+x[8])-4*(x[-12]+x[16])-x[-16]-x[20];
				 y[1]=47*(x[1]+x[5])+22*(x[-3]+x[9])-4*(x[-11]+x[17])-x[-15]-x[21];
				 y[2]=47*(x[2]+x[6])+22*(x[-2]+x[10])-4*(x[-10]+x[18])-x[-14]-x[22];
				 px=0;
				 x+=12;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_5_2_ver_0(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,int *x7,int *x8,long n)
{
	register long	a,b,c;

	do
	{
		a=52*x4[0]+36*(x3[0]+x5[0])+9*(x2[0]+x6[0])-4*(x1[0]+x7[0])-3*(x0[0]+x8[0]); sclip(a,14);
		b=52*x4[1]+36*(x3[1]+x5[1])+9*(x2[1]+x6[1])-4*(x1[1]+x7[1])-3*(x0[1]+x8[1]); sclip(b,14);
		c=52*x4[2]+36*(x3[2]+x5[2])+9*(x2[2]+x6[2])-4*(x1[2]+x7[2])-3*(x0[2]+x8[2]); sclip(c,14);

		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3; x7+=3; x8+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_5_2_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,int *x7,long n)
{
	register long	a,b,c;

	do
	{
		a=47*(x3[0]+x4[0])+22*(x2[0]+x5[0])-4*(x1[0]+x6[0])-x0[0]-x7[0]; sclip(a,14);
		b=47*(x3[1]+x4[1])+22*(x2[1]+x5[1])-4*(x1[1]+x6[1])-x0[1]-x7[1]; sclip(b,14);
		c=47*(x3[2]+x4[2])+22*(x2[2]+x5[2])-4*(x1[2]+x6[2])-x0[2]-x7[2]; sclip(c,14);

		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3; x7+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_5_2(int *tb0,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[10];
	register int	*tb1,*tbh;
	register long	i;

	dx*=4;
	px*=5;
	py*=5;
	x+=4*((px>>1)+border)+((py>>1)+border-4)*dx;
	px&=1;
	py&=1;

	for (i=0; i<8; i++)
	{
		buf[i]=tb0;
		scale_5_2_hor(tb0,x,px,w); x+=dx;
		tb0+=3*w;
	}
	buf[8]=tb0;	tb1=tb0+3*w;
	buf[9]=tb1;

	for (; h; h--)
	{
		scale_5_2_hor(tb0,x,px,w); x+=dx;
		scale_5_2_hor(tb1,x,px,w); x+=dx;
		
		switch ( py ) {
			case 0 : scale_5_2_ver_0(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],buf[7],buf[8],w);	y+=dy; py=1; 
				 tb0=buf[0];
				 tb1=buf[1];
				 for (i=0; i<8; i++) buf[i]=buf[i+2];
				 buf[8]=tb0;
				 buf[9]=tb1;
				 continue;

			case 1 : scale_5_2_ver_1(y,buf[0],buf[1],buf[3],buf[4],buf[5],buf[6],buf[8],buf[9],w);		y+=dy; py=0;
				 tbh=buf[0];
				 tb0=buf[1];
				 tb1=buf[2];
				 for (i=0; i<7; i++) buf[i]=buf[i+3];
				 buf[7]=tbh;
				 buf[8]=tb0;
				 buf[9]=tb1;
				 scale_5_2_hor(tbh,x,px,w); x+=dx;
				 continue;
		}
	}
}

static void scale_8_3_hor(int *y,uchar *x,long px,long n)
{
	for (; n; n--)
	{
		switch ( px )
		{
			case 0 :
				 y[0]=96*x[0]+70*(x[-4]+x[4])+21*(x[-8]+x[8])-5*(x[-12]+x[12])-6*(x[-16]+x[16]);
				 y[1]=96*x[1]+70*(x[-3]+x[5])+21*(x[-7]+x[9])-5*(x[-11]+x[13])-6*(x[-15]+x[17]);
				 y[2]=96*x[2]+70*(x[-2]+x[6])+21*(x[-6]+x[10])-5*(x[-10]+x[14])-6*(x[-14]+x[18]);
				 px=2;
				 x+=8;
				 y+=3;
				 continue;
			case 1 :
				 y[0]=-4*x[-16]-8*x[-12]+9*x[-8]+54*x[-4]+92*x[0]+83*x[4]+37*x[8]-7*x[16];
				 y[1]=-4*x[-15]-8*x[-11]+9*x[-7]+54*x[-3]+92*x[1]+83*x[5]+37*x[9]-7*x[17];
				 y[2]=-4*x[-14]-8*x[-10]+9*x[-6]+54*x[-2]+92*x[2]+83*x[6]+37*x[10]-7*x[18];
				 px=0;
				 x+=12;
				 y+=3;
				 continue;
			case 2 :
				 y[0]=-7*x[-12]+37*x[-4]+83*x[0]+92*x[4]+54*x[8]+9*x[12]-8*x[16]-4*x[20];
				 y[1]=-7*x[-11]+37*x[-3]+83*x[1]+92*x[5]+54*x[9]+9*x[13]-8*x[17]-4*x[21];
				 y[2]=-7*x[-10]+37*x[-2]+83*x[2]+92*x[6]+54*x[10]+9*x[14]-8*x[18]-4*x[22];
				 px=1;
				 x+=12;
				 y+=3;
				 continue;
		}
	}
}

static inline void scale_8_3_ver_0(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,int *x7,int *x8,long n)
{
	register long	a,b,c;

	do
	{
		a=96*x4[0]+70*(x3[0]+x5[0])+21*(x2[0]+x6[0])-5*(x1[0]+x7[0])-6*(x0[0]+x8[0]); sclip(a,16);
		b=96*x4[1]+70*(x3[1]+x5[1])+21*(x2[1]+x6[1])-5*(x1[1]+x7[1])-6*(x0[1]+x8[1]); sclip(b,16);
		c=96*x4[2]+70*(x3[2]+x5[2])+21*(x2[2]+x6[2])-5*(x1[2]+x7[2])-6*(x0[2]+x8[2]); sclip(c,16);

		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3; x7+=3; x8+=3;
		y+=4;
		n--;
	}
	while ( n );
}

static inline void scale_8_3_ver_1(uchar *y,int *x0,int *x1,int *x2,int *x3,int *x4,int *x5,int *x6,int *x7,long n)
{
	register long	a,b,c;

	do
	{
		a=-4*x0[0]-8*x1[0]+9*x2[0]+54*x3[0]+92*x4[0]+83*x5[0]+37*x6[0]-7*x7[0]; sclip(a,16);
		b=-4*x0[1]-8*x1[1]+9*x2[1]+54*x3[1]+92*x4[1]+83*x5[1]+37*x6[1]-7*x7[1]; sclip(b,16);
		c=-4*x0[2]-8*x1[2]+9*x2[2]+54*x3[2]+92*x4[2]+83*x5[2]+37*x6[2]-7*x7[2]; sclip(c,16);

		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=3; x1+=3; x2+=3; x3+=3; x4+=3; x5+=3; x6+=3; x7+=3;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_8_3(int *tb0,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	int		*buf[9];
	register int	*tb1,*tbh;
	register long	i;

	dx*=4;
	px*=8;
	py*=8;
	x+=4*((px/3)+border)+((py/3)+border-4)*dx;
	px%=3;
	py%=3;

	for (i=0; i<7; i++)
	{
		buf[i]=tb0;
		scale_8_3_hor(tb0,x,px,w); x+=dx;
		tb0+=3*w;
	}
	tb1=tb0+3*w;

	for (; h; h--)
	{
		buf[7]=tb0;
		buf[8]=tb1;

		scale_8_3_hor(tb0,x,px,w); x+=dx;
		scale_8_3_hor(tb1,x,px,w); x+=dx;

		switch ( py ) {
			case 0 : scale_8_3_ver_0(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],buf[7],buf[8],w);	y+=dy; py=2;
				 tb0=buf[0];
				 tb1=buf[1];
				 for (i=0; i<7; i++) buf[i]=buf[i+2];
				 continue;

			case 1 : scale_8_3_ver_1(y,buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],buf[8],w);		y+=dy; py=0;

			case 2 : tbh=buf[0];
				 tb0=buf[1];
				 tb1=buf[2];
				 for (i=0; i<6; i++) buf[i]=buf[i+3];
				 buf[6]=tbh;
				 scale_8_3_hor(tbh,x,px,w); x+=dx;
				 if ( !py ) continue;

				 scale_8_3_ver_1(y,tbh,buf[5],buf[4],buf[3],buf[2],buf[1],buf[0],tb0,w);		y+=dy; py=1;
				 continue;
		}
	}
}


static void scale_3_1_hor(int *y,uchar *x,long n)
{
	do
	{
		y[0]=86*x[0]+68*(x[-4]+x[4])+31*(x[-8]+x[8])-10*(x[-16]+x[16])-4*(x[-20]+x[20]);
		y[1]=86*x[1]+68*(x[-3]+x[5])+31*(x[-7]+x[9])-10*(x[-15]+x[17])-4*(x[-19]+x[21]);
		y[2]=86*x[2]+68*(x[-2]+x[6])+31*(x[-6]+x[10])-10*(x[-14]+x[18])-4*(x[-18]+x[22]);
		x+=12;
		y+=9;
		n--;
	}
	while ( n );
}

static inline void scale_3_1_ver(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=86*x1[6]+68*(x1[3]+x2[0])+31*(x1[0]+x2[3])-10*(x0[3]+x3[0])-4*(x0[0]+x3[3]); sclip(a,16);
		b=86*x1[7]+68*(x1[4]+x2[1])+31*(x1[1]+x2[4])-10*(x0[4]+x3[1])-4*(x0[1]+x3[4]); sclip(b,16);
		c=86*x1[8]+68*(x1[5]+x2[2])+31*(x1[2]+x2[5])-10*(x0[5]+x3[2])-4*(x0[2]+x3[5]); sclip(c,16);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=9; x1+=9; x2+=9; x3+=9;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_3_1(int *tb0,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	register int	*tb1,*tb2,*tb3, *ttb;
	register long	i;

	dx*=4;
	x+=4*(3*px+border)+(3*py+border-5)*dx;

	tb1=tb0+9*w;	for (i=0; i<3; i++) { scale_3_1_hor(tb0+3*i,x,w); x+=dx; }
	tb2=tb1+9*w;	for (i=0; i<3; i++) { scale_3_1_hor(tb1+3*i,x,w); x+=dx; }
	tb3=tb2+9*w;	for (i=0; i<3; i++) { scale_3_1_hor(tb2+3*i,x,w); x+=dx; }

	do
	{
		for (i=0; i<3; i++) { scale_3_1_hor(tb3+3*i,x,w); x+=dx; }

		scale_3_1_ver(y,tb0,tb1,tb2,tb3,w); y+=dy;

		ttb=tb0;
		tb0=tb1; tb1=tb2; tb2=tb3; tb3=ttb;

		h--;
	}
	while ( h );
}

static void scale_4_1_hor(int *y,uchar *x,long n)
{
	do
	{
		y[0]=32*x[0]+28*(x[-4]+x[4])+19*(x[-8]+x[8])+8*(x[-12]+x[12])-3*(x[-20]+x[20]+x[-24]+x[24])-(x[-28]+x[28]);
		y[1]=32*x[1]+28*(x[-3]+x[5])+19*(x[-7]+x[9])+8*(x[-11]+x[13])-3*(x[-19]+x[21]+x[-23]+x[25])-(x[-27]+x[29]);
		y[2]=32*x[2]+28*(x[-2]+x[6])+19*(x[-6]+x[10])+8*(x[-10]+x[14])-3*(x[-18]+x[22]+x[-22]+x[26])-(x[-26]+x[30]);
		x+=16;
		y+=12;
		n--;
	}
	while ( n );
}

static inline void scale_4_1_ver(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=32*x1[ 9]+28*(x1[6]+x2[0])+19*(x1[3]+x2[3])+8*(x1[0]+x2[6])-3*(x0[3]+x0[6]+x3[0]+x3[3])-x0[0]-x3[6]; sclip(a,14);
		b=32*x1[10]+28*(x1[7]+x2[1])+19*(x1[4]+x2[4])+8*(x1[1]+x2[7])-3*(x0[4]+x0[7]+x3[1]+x3[4])-x0[1]-x3[7]; sclip(b,14);
		c=32*x1[11]+28*(x1[8]+x2[2])+19*(x1[5]+x2[5])+8*(x1[2]+x2[8])-3*(x0[5]+x0[8]+x3[2]+x3[5])-x0[2]-x3[8]; sclip(c,14);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=12; x1+=12; x2+=12; x3+=12;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_4_1(int *tb0,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	register int	*tb1,*tb2,*tb3,*ttb;
	register long	i;

	dx*=4;
	x+=4*(4*px+border)+(4*py+border-7)*dx;

	tb1=tb0+12*w;	for (i=0; i<4; i++) { scale_4_1_hor(tb0+3*i,x,w); x+=dx; }
	tb2=tb1+12*w;	for (i=0; i<4; i++) { scale_4_1_hor(tb1+3*i,x,w); x+=dx; }
	tb3=tb2+12*w;	for (i=0; i<4; i++) { scale_4_1_hor(tb2+3*i,x,w); x+=dx; }

	do
	{
		for (i=0; i<4; i++) { scale_4_1_hor(tb3+3*i,x,w); x+=dx; }

		scale_4_1_ver(y,tb0,tb1,tb2,tb3,w); y+=dy;

		ttb=tb0;
		tb0=tb1; tb1=tb2; tb2=tb3; tb3=ttb;

		h--;
	}
	while ( h );
}

static void scale_5_1_hor(int *y,uchar *x,long n)
{
	do
	{
		y[0]=50*x[0]+46*(x[-4]+x[4])+35*(x[-8]+x[8])+21*(x[-12]+x[12])+8*(x[-16]+x[16])-3*(x[-24]+x[24]+x[-28]+x[28])-(x[-32]+x[32]);
		y[1]=50*x[1]+46*(x[-3]+x[5])+35*(x[-7]+x[9])+21*(x[-11]+x[13])+8*(x[-15]+x[17])-3*(x[-23]+x[25]+x[-27]+x[29])-(x[-31]+x[33]);
		y[2]=50*x[2]+46*(x[-2]+x[6])+35*(x[-6]+x[10])+21*(x[-10]+x[14])+8*(x[-14]+x[18])-3*(x[-22]+x[26]+x[-26]+x[30])-(x[-30]+x[34]);
		x+=20;
		y+=15;
		n--;
	}
	while ( n );
}

static inline void scale_5_1_ver(uchar *y,int *x0,int *x1,int *x2,int *x3,long n)
{
	register long	a,b,c;

	do
	{
		a=50*x1[ 9]+46*(x1[6]+x1[12])+35*(x1[3]+x2[0])+21*(x1[0]+x2[3])+8*(x0[12]+x2[6])-3*(x0[3]+x0[6]+x2[12]+x3[0])-x0[0]-x3[3]; sclip(a,16);
		b=50*x1[10]+46*(x1[7]+x1[13])+35*(x1[4]+x2[1])+21*(x1[1]+x2[4])+8*(x0[13]+x2[7])-3*(x0[4]+x0[7]+x2[13]+x3[1])-x0[1]-x3[4]; sclip(b,16);
		c=50*x1[11]+46*(x1[8]+x1[14])+35*(x1[5]+x2[2])+21*(x1[2]+x2[5])+8*(x0[14]+x2[8])-3*(x0[5]+x0[8]+x2[14]+x3[2])-x0[2]-x3[5]; sclip(c,16);
		y[0]=a; y[1]=b; y[2]=c; y[3]=0;
		x0+=15; x1+=15; x2+=15; x3+=15;
		y+=4;
		n--;
	}
	while ( n );
}

void scale_5_1(int *tb0,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h)
{
	register int	*tb1,*tb2,*tb3,*ttb;
	register long	i;

	dx*=4;
	x+=4*(5*px+border)+(5*py+border-7)*dx;

	tb1=tb0+15*w;	for (i=0; i<5; i++) { scale_5_1_hor(tb0+3*i,x,w); x+=dx; }
	tb2=tb1+15*w;	for (i=0; i<5; i++) { scale_5_1_hor(tb1+3*i,x,w); x+=dx; }
	tb3=tb2+15*w;	for (i=0; i<5; i++) { scale_5_1_hor(tb2+3*i,x,w); x+=dx; }

	do
	{
		for (i=0; i<5; i++) { scale_5_1_hor(tb3+3*i,x,w); x+=dx; }

		scale_5_1_ver(y, tb0, tb1, tb2,	tb3, w); y+=dy;

		ttb=tb0;
		tb0=tb1; tb1=tb2; tb2=tb3; tb3=ttb;
		h--;
	}
	while ( h );
}


#ifdef test

#define rdtsc(low,high) __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
#define rdtscl(low)     __asm__ __volatile__("rdtsc" : "=a" (low) : : "edx")

#define n	400

int main()
{
	long	a,b,c,i;
	int	x[32+1200];
	int	y[n+4];

	for (i=0; i<10; i++)
	{
		rdtscl(a);
		rdtscl(b);
		scale_3_1_hor(y  ,(uchar *) (x+16),n/3);
		scale_3_1_hor(y+1,(uchar *) (x+16),n/3);
		scale_3_1_hor(y+2,(uchar *) (x+16),n/3);
		rdtscl(c);
	}
	printf("%f %f\n",((double) (b-a))/((double) n),((double) (c-b))/((double) n));
	return 0;
}
#endif
