
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include <png.h>

#include "sc.h"

#include "jhead.h"

long load_jpeg(long *par,long fd);
extern uchar msg_error[];
extern uchar msg_deleted[];

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/XShm.h>
#include <X11/keysym.h>

#define panstep		64

#define numsc		25
#define numnmsc		16

#define sc_half		15
#define sc_full		20
#define sc_double	22
#define sc_msg		21

#define infoheight	16
#define infotx		20
#define infoty		3
#define infoscw		100

#define maxnamelen	192

#define flag_deleted	1
#define flag_tagged	2


#define min(a,b)	(((a)<(b))?(a):(b));
#define clip(x)		if ( x<0 ) x=0; if ( x>255 ) x=255
#define sclip(x,s)	x=(x+(1<<(s-1)))>>s; if ( x<0 ) x=0; if ( x>255 ) x=255

typedef struct {
	void (* f)(int *tb,uchar *y,long dy,uchar *x,long dx,long px,long py,long w,long h);
	int	n,z,mip;
} Tsc;

typedef struct {
	uchar	*buf;
	uchar	*mbuf;
	char	name[maxnamelen];
	char	*typ;
	long	bsize;
	int	maxsc;
	int	ix,iy;
	int	dx,dy;
	int	mx,my;
} Timg;

typedef struct {
	char	*name;
	int	flags,x0,y0;
} Tinfo;

static Display		*disp;
static Colormap		cmap;
static int		shift_mask,alt_mask,ctrl_mask,caps_lock,num_lock,scroll_lock;
static	Tsc		scale3 = {scale_3_1,3,1,0};


static long skip_comment(FILE *f)
{
	register long	i;

	i=getc(f);

	for (;;)
	{
		while ( i<=32 ) i=getc(f);
		if ( i!='#' ) break;

		do i=getc(f); while ( i!=10 );
		i=getc(f);
		if ( i==13 ) i=getc(f);
	}

	return i;
}

static long load_pnm(Timg *img,FILE *f,long a,long b,long c)
{
	static char	*ppmtyp[4] = {"PGM-8bit","PGM-16bit","PPM-24bit","PPM-48bit"};
	register long	i,j,k,dx,dy,mode,bpp;
	register uchar	*buf,*q,*p,*r;

	if ( b=='5' ) mode=0; else
	if ( b=='6' ) mode=2; else return -1;

	a=skip_comment(f)-'0';

	for (;;)
	{
		b=getc(f)-'0';
		if ( b<0 || (b>9) ) break;
		a=a*10+b;
	}
	do b=getc(f)-'0'; while ( (b<0) || (b>9) );
	for (;;)
	{
		c=getc(f)-'0';
		if ( c<0 || (c>9) ) break;
		b=b*10+c;
	}

	dx=a+2*border;
	dy=b+2*border;

	a=skip_comment(f)-'0';
	for (;;)
	{
		b=getc(f)-'0';
		if ( b<0 || (b>9) ) break;
		a=a*10+b;
	}
	b=0;
	q=NULL;
	if ( a>256 ) mode++;
	bpp=mode+1;
	if ( mode==3 )
	{
		q=(uchar *) alloca(6*(dx-2*border));
		bpp=6;
	}

	buf=img->buf;
	i=4*dx*dy;
	j=img->bsize;
	if ( buf && (i>j) ) buf=realloc(buf,j=i);
	if ( !buf ) buf=malloc(j=i);

	img->bsize=j;
	img->buf=buf;
	img->ix=dx;
	img->iy=dy;
	img->typ=ppmtyp[mode];

	p=buf+4*border+4*dx*border;
	i=bpp*(dx-2*border)*(dy-2*border);

	if ( mode<3 )
	{
		q=buf+4*dx*dy-4*border*dx-i;
		fread(q,1,i,f);
	}

	k=4*(dx-2*border);

	for (i=0; i<dy-2*border; i++)
	{
		switch ( mode )
		{
			case 0 : for (j=0; j<k; j+=4) { p[j]=p[j+1]=p[j+2]=q[0]; p[j+3]=0; q++;  }		break;
			case 1 : for (j=0; j<k; j+=4) { p[j]=p[j+1]=p[j+2]=q[1]; p[j+3]=0; q+=2; }		break;
			case 2 : for (j=0; j<k; j+=4) { p[j]=q[2]; p[j+1]=q[1]; p[j+2]=q[0]; p[j+3]=0; q+=3; }	break;
			case 3 :
				fread(q,1,6*(dx-2*border),f);
				for (j=0,r=q; j<k; j+=4,r+=6)
				{
					p[j  ]=r[5];
					p[j+1]=r[3];
					p[j+2]=r[1];
					p[j+3]=0;
				}
				break;
		}

		for (j=0; j<4*border; j+=4)
		{
			*((uint *) (p-4-j))=*((uint *) (p+4+j));
			*((uint *) (p+4*dx-8*border+j))=*((uint *) (p+4*dx-8*border-8-j));
		}
		if ( (i>0) && (i<=border) ) memcpy(p-4*border-8*i*dx,p-4*border, 4*dx);
		j=(dy-2*border-1)-i;
		if ( (j>0) && (j<=border) ) memcpy(p-4*border+8*j*dx,p-4*border, 4*dx);

		p+=4*dx;
	}
	return 0;
}

static long load_tga(Timg *img,FILE *f,long c0,long c1,long c2)
{
	uchar		hd[18];
	register long	i,j,k,dx,dy;
	register uchar	*buf,*p,*q;

	fread(hd+3,1,15,f);

	dx=((long) hd[12]) + (((long) hd[13]) << 8) + 2*border;
	dy=((long) hd[14]) + (((long) hd[15]) << 8) + 2*border;

	while ( c0-- ) getc(f);

	if ( (hd[16]>>3)!=3 )
	{
		fprintf(stderr,"Only support RGB-TARGAs\n");
		return -1;
	}

	buf=img->buf;
	i=4*dx*dy;
	j=img->bsize;
	if ( buf && (i>j) ) buf=realloc(buf,j=i);
	if ( !buf ) buf=malloc(j=i);

	img->bsize=j;
	img->buf=buf;
	img->ix=dx;
	img->iy=dy;
	img->typ="TARGA";

	p=buf+4*border+4*dx*border;
	k=dx-2*border;

	for (i=0; i<dy-2*border; i++)
	{
		q=p+k;
		fread(q,1,3*k,f);
		for (j=0; j<4*k; j+=4)
		{
			p[j  ]=q[0];
			p[j+1]=q[1];
			p[j+2]=q[2];
			p[j+3]=0;
			q+=3;
		}

		for (j=0; j<4*border; j+=4)
		{
			*((uint *) (p-4-j))=*((uint *) (p+4+j));
			*((uint *) (p+4*dx-8*border+j))=*((uint *) (p+4*dx-8*border-8-j));
		}
		if ( (i>0) && (i<=border) ) memcpy(p-4*border-8*i*dx,p-4*border, 4*dx);
		j=(dy-2*border-1)-i;
		if ( (j>0) && (j<=border) ) memcpy(p-4*border+8*j*dx,p-4*border, 4*dx);

		p+=4*dx;
	}
	return 0;
}

static long load_jpg(Timg *img,FILE *f,long c0,long c1,long c2)
{
	long	jp[5];
	long	fd = fileno(f);

	lseek(fd,0,SEEK_SET);

	jp[0]=(long) img->buf;
	jp[1]=(long) img->bsize;

	if ( load_jpeg(jp,fd)<0 ) return -1;

	img->buf  =(uchar *) jp[0];
	img->bsize=jp[1];
	img->ix=jp[2]+2*border;
	img->iy=jp[3]+2*border;
	img->typ=(char *) jp[4];
	return 0;
}

static long load_msg(Timg *img,uchar *msg,char *typ)
{
	long	jp[5];

	jp[0]=(long) img->buf;
	jp[1]=(long) img->bsize;
	jp[2]=(long) msg;

	if ( load_jpeg(jp,-1)<0 )
	{
		fprintf(stderr,"Error whie decoding JPEG message\n");
		exit(0);
	}

	img->buf  =(uchar *) jp[0];
	img->bsize=jp[1];
	img->ix=jp[2]+2*border;
	img->iy=jp[3]+2*border;
	img->typ=typ;
	img->maxsc=sc_msg;

	return 0;
}

static long load_png(Timg *img,FILE *f,long c0,long c1,long c2)
{
	static char	pngtyp[32];
	png_structp	png_ptr;
	png_infop	info_ptr;
	png_bytep	row_pointers[1];
	register long	i,j,bpp;
	png_uint_32	dx,dy;
	int		bit_depth, color_type, interlace_type;
	uchar		*buf, *p;

	ungetc(c2,f);
	ungetc(c1,f);
	ungetc(c0,f);

	png_ptr = png_create_read_struct(
			PNG_LIBPNG_VER_STRING,
			NULL,				/* (png_voidp) user_error_ptr */
			NULL, 				/* user_error_fn */
			NULL				/* user_warning_fn */
			);

	info_ptr = png_create_info_struct(png_ptr);
	setjmp(png_jmpbuf(png_ptr));
	png_init_io(png_ptr, f);
	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &dx, &dy, &bit_depth, &color_type, &interlace_type, NULL, NULL);

	bpp=bit_depth;
	if ( bit_depth==16 ) png_set_strip_16(png_ptr);
	if ( color_type==PNG_COLOR_TYPE_PALETTE ) png_set_palette_to_rgb(png_ptr);
	if ( color_type==PNG_COLOR_TYPE_GRAY || color_type==PNG_COLOR_TYPE_GRAY_ALPHA ) png_set_gray_to_rgb(png_ptr);
	if ( color_type & PNG_COLOR_MASK_ALPHA ) png_set_strip_alpha(png_ptr);
	if ( bit_depth<8 ) png_set_packing(png_ptr);
	if ( color_type==PNG_COLOR_TYPE_RGB || color_type==PNG_COLOR_TYPE_RGB_ALPHA ) bpp*=3;
	png_set_bgr(png_ptr);
	png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);


	dx+=2*border;
	dy+=2*border;

	buf=img->buf;
	i=4*dx*dy;
	j=img->bsize;
	if ( buf && (i>j) ) buf=realloc(buf,j=i);
	if ( !buf ) buf=malloc(j=i);

	img->bsize=j;
	img->buf=buf;
	img->ix=dx;
	img->iy=dy;
	img->typ=pngtyp;
	sprintf(pngtyp,"PNG-%libit",bpp);

	p=buf+4*border+4*dx*border;

	for (i=0; i<(long) (dy-2*border); i++)
	{
		row_pointers[0]=(png_bytep) p;
		png_read_rows(png_ptr, row_pointers, NULL, 1);

		for (j=0; j<4*border; j+=4)
		{
			*((uint *) (p-4-j))=*((uint *) (p+4+j));
			*((uint *) (p+4*dx-8*border+j))=*((uint *) (p+4*dx-8*border-8-j));
		}
		if ( (i>0) && (i<=border) ) memcpy(p-4*border-8*i*dx,p-4*border, 4*dx);
		j=(dy-2*border-1)-i;
		if ( (j>0) && (j<=border) ) memcpy(p-4*border+8*j*dx,p-4*border, 4*dx);

		p+=4*dx;
	}

	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

	return 0;
}


static inline long is_pnm(long a,long b,long c)
{
	return (a=='P') && (b>='4') && (b<='6') && ((c==8)||(c==10)||(c==32));
}

static inline long is_tga(long a,long b,long c)
{
	return (b<2) && (c>=2) && (c<=3);
}

static inline long is_jpg(long a,long b,long c)
{
	return (a==0xFF) && (b==0xD8);
}

static inline long is_png(long a,long b,long c)
{
	return (a==137) && (b==80) && (c==78);
}

static void load_image(Timg *img,Tinfo *ifile)
{
	FILE		*f;
	register long	r,c0,c1,c2;

	for (r=0; (r<maxnamelen-1) && (c0=ifile->name[r]); r++) img->name[r]=(uchar) c0;
	img->name[r]=0;
	r=-1;

	if ( ifile->flags & flag_deleted ) r=load_msg(img,msg_deleted,"DELETED"); else
	{
		if ( (ifile->name[0]!='-') || ifile->name[1] ) f=fopen(ifile->name,"r"); else f=stdin;

		if ( f )
		{

			c0=getc(f);
			c1=getc(f);
			c2=getc(f);

			if ( is_pnm(c0,c1,c2) ) r=load_pnm(img,f,c0,c1,c2); else
			if ( is_tga(c0,c1,c2) ) r=load_tga(img,f,c0,c1,c2); else
			if ( is_png(c0,c1,c2) ) r=load_png(img,f,c0,c1,c2); else
			if ( is_jpg(c0,c1,c2) ) r=load_jpg(img,f,c0,c1,c2);

			img->maxsc=numsc-1;
		}

		if ( r<0 ) load_msg(img,msg_error,"ERROR");
		if ( f && (f!=stdin) ) fclose(f);
	}
	img->mbuf=NULL;
}

static void rrot(Timg *img)
{
	register uint	*x,*y,*p;
	register long	s,w,h,i,j;

	x=(uint *) img->buf;
	w=img->ix;
	h=img->iy;
	s=w*h;
	img->buf=(uchar *) (y=(uint *) malloc(4*s));
	img->bsize=4*s;

	for (i=0; i<w; i++)
	{
		p=x+s+i;

		for (j=h; j; j--)
		{
			p-=w;
			*(y++)=*p;
		}
	}

	free(x);
	img->ix=h;
	img->iy=w;
	img->mbuf=NULL;
}

static void lrot(Timg *img)
{
	register uint	*x,*y,*p;
	register long	s,w,h,i,j;

	x=(uint *) img->buf;
	w=img->ix;
	h=img->iy;
	s=w*h;
	img->buf=(uchar *) (y=(uint *) malloc(4*s));
	img->bsize=4*s;

	for (i=w-1; i>=0; i--)
	{
		p=x+i;

		for (j=h; j; j--)
		{
			*(y++)=*p;
			p+=w;
		}
	}

	free(x);
	img->ix=h;
	img->iy=w;
	img->mbuf=NULL;
}

static void rot180(Timg *img)
{
	register uint	*x,*y;
	register long	w,h,i,j,a,b;

	x=(uint *) img->buf;
	w=img->ix;
	h=img->iy;
	y=x+w*h;

	for (i=h>>1; i; i--)
	{
		for (j=w; j; j--)
		{
			y--;
			a=*x; b=*y;
			*x=b; *y=a;
			x++;
		}
	}
	if ( h&1 )
	{
		for (j=w>>1; j; j--)
		{
			y--;
			a=*x; b=*y;
			*x=b; *y=a;
			x++;
		}
	}
	img->mbuf=NULL;
}

static void mirror(Timg *img)
{
	register uint	*x0,*x,*y;
	register long	w,h,j,a,b;

	x0=(uint *) img->buf;
	w=img->ix;
	h=img->iy;

	while ( h-- )
	{
		x=x0;
		y=x0+w;
		for (j=w>>1; j; j--)
		{
			y--;
			a=*x; b=*y;
			*x=b; *y=a;
			x++;
		}
		x0+=w;
	}
	img->mbuf=NULL;
}

static void convert16(uchar *y,uchar *x,long dx,long w,long h)
{
	register long	n;

	dx-=w;
	do
	{
		n=w;
		do
		{
			y[0] = (uchar) ((x[0]>>3) | ((x[1]<<3)&(128+64+32     )));
			y[1] = (uchar) ((x[1]>>5) | ( x[2]    &(128+64+32+16+8)));
			y+=2;
			x+=4;
			n--;
		}
		while ( n );

		x+=4*dx;
		y+=2*dx;
		h--;
	}
	while ( h );
}

static void make_mip(Timg *img,Tsc *sc)
{
	register long	i,j,w,h,ix,iy,mx,my;
	register uchar	*buf,*mbuf,*p;
	register int	*tbuf;

	ix=img->ix;
	iy=img->iy;

	buf=img->buf;

	mx=(2*sc->z*(ix-2*border)+sc->n)/(2*sc->n) + 2*border;
	my=(2*sc->z*(iy-2*border)+sc->n)/(2*sc->n) + 2*border;

	i=  4*ix*iy;
	j=i+4*mx*my;

	if ( j>img->bsize ) img->buf=buf=realloc(buf,img->bsize=j);

	mbuf=buf+i;
	img->mbuf=mbuf;
	img->mx=mx;
	img->my=my;

	p=mbuf+4*(border*mx+border);

	w=mx-2*border;
	h=my-2*border;

	tbuf=(int *) alloca(15*4*w*sizeof(int)+8);
	tbuf=(int *) roundup((unsigned long) tbuf);

	sc->f(tbuf,p,4*mx,buf,ix,0,0,w,h);

	for (i=0; i<h; i++)
	{
		for (j=0; j<4*border; j+=4)
		{
			*((uint *) (p-4-j))=*((uint *) (p+4+j));
			*((uint *) (p+4*mx-8*border+j))=*((uint *) (p+4*mx-8*border-8-j));
		}
		if ( (i>0) && (i<=border) ) memcpy(p-4*border-8*i*mx,p-4*border, 4*mx);
		j=(my-2*border-1)-i;
		if ( (j>0) && (j<=border) ) memcpy(p-4*border+8*j*mx,p-4*border, 4*mx);

		p+=4*mx;
	}
}

static void write_mip_ppm(Timg *img)
{
	register uchar	*buf,*x,*p,*q;
	register long	r,g,b,i,d,w,h;

	d=img->mx;
	h=img->my-2*border;
	w=d-2*border;
	x=img->mbuf+4*(border*d+border);

	buf=(uchar *) alloca(3*w);

	printf("P6\n%li %li\n255\n",w,h);

	while ( h-- )
	{
		p=buf;
		q=x;
		for (i=w; i; i--)
		{
			b=q[0]; g=q[1]; r=q[2];
			p[0]=(uchar) r; p[1]=(uchar) g; p[2]=(uchar) b;
			q+=4;
			p+=3;
		}
		fwrite(buf,3,w,stdout);
		x+=4*d;
	}
}

static void write_mip_tga(Timg *img)
{
	static uchar    hd[18] = {0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,24,32};
	register uchar	*buf,*x,*p,*q;
	register long	r,g,b,i,d,w,h;

	d=img->mx;
	h=img->my-2*border;
	w=d-2*border;
	x=img->mbuf+4*(border*d+border);

	hd[12]=(uchar) (w & 0xFF);
	hd[13]=(uchar) ((w >> 8 ) & 0xFF);
	hd[14]=(uchar) (h & 0xFF);
	hd[15]=(uchar) ((h >> 8 ) & 0xFF);

	fwrite(hd,1,18,stdout);

	buf=(uchar *) alloca(3*w);

	while ( h-- )
	{
		p=buf;
		q=x;
		for (i=w; i; i--)
		{
			b=q[0]; g=q[1]; r=q[2];
			p[0]=(uchar) b; p[1]=(uchar) g; p[2]=(uchar) r;
			q+=4;
			p+=3;
		}
		fwrite(buf,3,w,stdout);
		x+=4*d;
	}
}

static void write_mip_png(Timg *img)
{
	register uchar		*buf,*x,*p,*q;
	register long		r,g,b,i,d,w,h;
	register png_structp	png_ptr;
	register png_infop	info_ptr;
	png_bytep		row_pointers[1];

	d=img->mx;
	h=img->my-2*border;
	w=d-2*border;
	x=img->mbuf+4*(border*d+border);

	buf=(uchar *) alloca(3*w);

	row_pointers[0]=buf;

	png_ptr = png_create_write_struct(
			PNG_LIBPNG_VER_STRING,
			NULL,				/* (png_voidp) user_error_ptr */
			NULL, 				/* user_error_fn */
			NULL				/* user_warning_fn */
			);

	info_ptr = png_create_info_struct(png_ptr);
	setjmp(png_jmpbuf(png_ptr));
	png_init_io(png_ptr, stdout);

	png_set_compression_level(png_ptr, 4);

	png_set_IHDR(png_ptr,info_ptr,w,h,8,
			PNG_COLOR_TYPE_RGB,
			PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT,
			PNG_FILTER_TYPE_DEFAULT
			);
	png_write_info(png_ptr, info_ptr);

	while ( h-- )
	{
		p=buf;
		q=x;
		for (i=w; i; i--)
		{
			b=q[0]; g=q[1]; r=q[2];
			p[0]=(uchar) r; p[1]=(uchar) g; p[2]=(uchar) b;
			q+=4;
			p+=3;
		}
		png_write_rows(png_ptr, row_pointers, 1);
		x+=4*d;
	}
	png_write_end(png_ptr, info_ptr);
}

static void usage(char *pn)
{
		fprintf(stderr, "usage: %s [-s[t|p] x y] <images (jpg,tga,png or ppm-raw)>\n"
			"options:\n"
			"-s[t|p] x y - scale image by x/y and write to stdout as PPM()/TGA(t)/PNG(p)\n"
			"hotkeys:\n"
			"  SPACE    - go forward\n"
			"  BACKSP.  - go backward (or b)\n"
			"  DEL      - toggle delete mark (delete marked files on exit)\n"
			"  Cursor   - move image around\n"
			"  Scr-Lock - invert cursor\n"
			"  home,end - go to first, last\n"
			"   c       - center image\n"
			"   g       - go to prefix\n"
			"  +, -     - zoom in, out\n"
			"  h,f,d    - zoom to half, full, double\n"
			"   s       - scale to max. size\n"
			"   t       - tag (untag) image (print tagged files on exit)\n"
			"  r,l      - rotate 90 deg. right, left\n"
			"   u       - rotate 180 deg.\n"
			"   m       - mirror\n"
			"   i       - toggle info line\n"
			"  0..9     - prefix number\n"
			"  ESC      - exit (no deletions)\n"
			"   q       - exit (+delete marked files)\n"
			"   z       - move window to top left\n",pn);
	exit(0);
}

static void get_xmod(Display *disp)
{
	register long			i,j,m,max;
	register XModifierKeymap	*mkm;

	shift_mask=0;
	alt_mask=0;
	ctrl_mask=0;
	caps_lock=0;
	num_lock=0;
	scroll_lock=0;

	mkm = XGetModifierMapping(disp);
	max = mkm->max_keypermod;

	for (i=0,m=1; m<256; i+=max,m<<=1)
	{
		for (j=i; j<i+max; j++)
		{
			switch ( XKeycodeToKeysym(disp, mkm->modifiermap[j], 0) )
			{
				case XK_Alt_L       :
				case XK_Alt_R       : alt_mask    |= m; break;
				case XK_Control_L   :
				case XK_Control_R   : ctrl_mask   |= m; break;
				case XK_Shift_L     :
				case XK_Shift_R     : shift_mask  |= m; break;
				case XK_Caps_Lock   : caps_lock   |= m; break;
				case XK_Scroll_Lock : scroll_lock |= m; break;
				case XK_Num_Lock    : num_lock    |= m; break;
				default             : break;
			}
		}
	}
	XFreeModifiermap(mkm);
}

static void get_area(Tsc *sc,Timg *img,int *tbuf,uchar *buffer,uchar *xbuf,long sx,long x0,long y0,long vw,long vh)
{
	if ( sc->mip )
	{
		if ( !img->mbuf ) make_mip(img,&scale3);

		sc->f(tbuf,buffer,4*sx,img->mbuf,img->mx,x0,y0,vw,vh);
	}
	else sc->f(tbuf,buffer,4*sx,img->buf,img->ix,x0,y0,vw,vh);

#ifdef big_endian
	if ( !xbuf )
	{
		long i;
		uchar *p = buffer;
		for (i=vh; i; i--,p+=4*sx) { memmove(p+1,p,4*vw-1); *p=0; }
	}
#endif

	if ( xbuf ) convert16( xbuf, buffer, sx, vw, vh );
}

int main(int argc,char **argv)
{
	Tsc		nmscaler[numnmsc] = {
				{scale_5_1, 5,1,0},
				{scale_4_1, 4,1,0},
				{scale_3_1, 3,1,0},
				{scale_8_3, 8,3,0},
				{scale_5_2, 5,2,0},
				{scale_9_4, 9,4,0},
				{scale_2_1, 2,1,0},
				{scale_5_3, 5,3,0},
				{scale_3_2, 3,2,0},
				{scale_4_3, 4,3,0},
				{scale_6_5, 6,5,0},
				{scale_1_1, 1,1,0},
				{scale_2_3, 2,3,0},
				{scale_1_2, 1,2,0},
				{scale_1_3, 1,3,0},
				{scale_1_4, 1,4,0}
			};
	Tsc		scaler[numsc] = {
				{scale_5_1,15,1,1},
				{scale_4_1,12,1,1},
				{scale_3_1, 9,1,1},
				{scale_8_3, 8,1,1},
				{scale_5_2,15,2,1},
				{scale_9_4,27,4,1},
				{scale_2_1, 6,1,1},
				{scale_5_3, 5,1,1},
				{scale_3_2, 9,2,1},
				{scale_4_3, 4,1,1},
				{scale_6_5,18,5,1},
				{scale_1_1, 3,1,1},
				{scale_8_3, 8,3,0},
				{scale_5_2, 5,2,0},
				{scale_9_4, 9,4,0},
				{scale_2_1, 2,1,0},
				{scale_5_3, 5,3,0},
				{scale_3_2, 3,2,0},
				{scale_4_3, 4,3,0},
				{scale_6_5, 6,5,0},
				{scale_1_1, 1,1,0},
				{scale_2_3, 2,3,0},
				{scale_1_2, 1,2,0},
				{scale_1_3, 1,3,0},
				{scale_1_4, 1,4,0}
			};
	char		t[384];
	Tinfo		*files;
	Window		rwin,win,twin;
	int		scrn;
	int		wzpos[2];
	XSizeHints	hint;
	XSetWindowAttributes	att;
	unsigned long		mask,blk;
	long		bx,by,sx,sy,sy0,tx,ty,x0,y0,i,opn,pn,depth,bytepp,scn,nsc;
	long		mx,my,vh,vw,num,info,autoscale=1;
	long		x0min,x0max,y0min,y0max;
	XEvent		ev;
	GC		gcont;
	XImage		*ximg;
	Visual		*vis;
	XShmSegmentInfo	shminfo;
	struct shmid_ds	shmbuf;
	uchar		*buffer;
	int		*tbuf;
	Timg		img;
	KeySym		ks;


	if ( argc<2 ) usage(argv[0]);

	if ( !strcmp(argv[1],"-p") )
	{
		uchar	*tmp;
		buffer=malloc( 4*2032*1532 );
		for (i=0; i<4*2032*1532; i++) buffer[i]=(uchar) (i^(i>>8));
		tmp   =malloc( 4*400*300+16 );
		tbuf=(int *) alloca(4*20*400*sizeof(int)+8);
		tbuf=(int *) roundup((unsigned long) tbuf);
		for (i=scn=0; i<100; i++)
		{
			if ( scn==numsc ) scn= 0; else
			if ( scn==    3 ) scn=12;
			scaler[scn].f(tbuf,tmp,1200,buffer,2032,0,0,400,300);
			scn++;
		}
		exit(0);
	}

	if ( !strcmp(argv[1],"-h") ) usage(argv[0]);


	files=alloca((argc-1)*sizeof(Tinfo));
	for (i=1; i<argc; i++)
	{
		files[i-1].name=argv[i];
		files[i-1].flags=0;
		files[i-1].x0=0;
		files[i-1].y0=0;
	}

	img.buf=NULL;
	img.bsize=0;

	if ( !strcmp(argv[1],"-s") || !strcmp(argv[1],"-st") || !strcmp(argv[1],"-sp") )
	{
		tx=atoi(argv[2]);
		ty=atoi(argv[3]);
		for (scn=0; scn<numnmsc; scn++)
		{
			if ( (nmscaler[scn].z==tx) && (nmscaler[scn].n==ty) ) break;
		}
		if ( scn==numnmsc )
		{
			fprintf(stderr,"Scaling by %li/%li not supported\nPossible options: ",tx,ty);
			for (scn=0; scn<numnmsc; scn++) fprintf(stderr," %i/%i",nmscaler[scn].z,nmscaler[scn].n);
			fprintf(stderr,"\n");
			exit(0);
		}
		for (i=3; i<argc-1; i++)
		{
			load_image(&img,files+i);
			make_mip(&img,nmscaler+scn);
			if ( argv[1][2]=='t' ) write_mip_tga(&img); else
			if ( argv[1][2]=='p' ) write_mip_png(&img); else write_mip_ppm(&img);
		}
		exit(0);
	}

	hint.flags = 0;
	disp = XOpenDisplay("");
	scrn = DefaultScreen(disp);
	cmap = XDefaultColormap(disp, scrn);

	sx  = DisplayWidth(disp, scrn);
	sy0 = DisplayHeight(disp, scrn);
	sy=sy0;

	XSynchronize(disp,True);

	vis   = XDefaultVisual(disp,scrn);
	depth = XDefaultDepth(disp,scrn);

	get_xmod(disp);

	fprintf(stderr,"Screen size: %lix%lix%li bpp\n",sx,sy,depth);

	switch ( depth ) {
		case 16 : bytepp=2;	break;
		case 24 : bytepp=4;	break;
		default : fprintf(stderr,"%li bpp visual not supported\n",depth);
			  exit(1);
	}

	mask = CWBackPixel | CWOverrideRedirect | CWEventMask;

	blk = BlackPixel(disp, scrn);

	att.background_pixel = blk;
	att.override_redirect = True;
	att.event_mask = ButtonPressMask | KeyPressMask;

	rwin=XRootWindow(disp,scrn);

	if ( bytepp==1 )
	{
		cmap=XCreateColormap(disp, rwin, vis, AllocAll);
		att.colormap=cmap;
		mask|=CWColormap;
	}

	win = XCreateWindow(disp, rwin,0,0,sx,sy,0,depth,InputOutput,vis,mask,&att);

	XSelectInput(disp, win, KeyPressMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | ExposureMask);

	hint.flags = PSize | PPosition;
	hint.x = 0;
	hint.y = 0;
	hint.width = sx;
	hint.height = sy;
	XSetStandardProperties(disp, win, "IMGView", "no_icon", (Pixmap) 0, NULL, 0, &hint);

 	XMapRaised(disp, win);
/*
	ev.xclient.type = ClientMessage;
	ev.xclient.serial = 0;
	ev.xclient.send_event = True;
	ev.xclient.message_type = XInternAtom(disp, "_NET_WM_STATE", False);
	ev.xclient.window = win;
	ev.xclient.format = 32;
	ev.xclient.data.l[0] = 1;
	ev.xclient.data.l[1] = XInternAtom(disp,"_NET_WM_STATE_FULLSCREEN",False);
	ev.xclient.data.l[2] = 0;
	ev.xclient.data.l[3] = 0;
	ev.xclient.data.l[4] = 0;
	if ( !XSendEvent(disp, rwin, False, SubstructureRedirectMask | SubstructureNotifyMask, &ev) )
	{
		fprintf(stderr,"Fullscreen failed\n");
	}
*/
	XTranslateCoordinates(disp,win,rwin,0,0,wzpos,wzpos+1,&twin);
//	XMoveWindow(disp, win, -wzpos[0], -wzpos[1]);

	gcont=XCreateGC(disp, win, 0, 0);

	XSetState(disp, gcont, -1, 0, GXcopy, -1);

	ximg=XShmCreateImage(disp,vis,depth,ZPixmap,NULL,&shminfo,sx,sy); 
	if ( ximg==NULL )
	{
		fprintf(stderr,"XShmCreateImage failed!\n");
		exit(1);
	}
	shminfo.shmid=shmget(IPC_PRIVATE,bytepp*sx*sy+8*8,IPC_CREAT | 0777);
	if ( shminfo.shmid==-1 )
	{
		fprintf(stderr,"shmget failed!\n");
		exit(1);
	}
	shminfo.shmaddr = (char *) shmat(shminfo.shmid, 0, 0);
	if (shminfo.shmaddr == ((char *) -1))
	{
		fprintf(stderr,"shmat failed!\n");
		exit(1);
	}
	ximg->data = shminfo.shmaddr;
	if ( bytepp==4 ) buffer = (uchar *) ximg->data; else buffer=malloc(4*sx*sy);
	shminfo.readOnly = False;
	XShmAttach(disp, &shminfo);
	memset(buffer, 0, bytepp*sx*sy);

	XSetInputFocus(disp, win, RevertToParent, CurrentTime);
	if ( bytepp==1 ) XInstallColormap(disp,cmap);

	tbuf=(int *) alloca(3*20*(sx+4)*sizeof(int)+8);
	tbuf=(int *) roundup((unsigned long) tbuf);

	pn=opn=0;
	bx=by=0;
	x0=y0=0;
	scn=nsc=sc_full;
	info=infoheight;
	sy=sy0-info;
	autoscale=1;

	for (;;)
	{

	reload:
		files[opn].x0=x0;
		files[opn].y0=y0;
		x0=files[pn].x0;
		y0=files[pn].y0;
		opn=pn;
		load_image(&img,files+pn);

	rescale:
		if ( autoscale )
		{
			for (scn=img.maxsc; scn>0; scn--)
			{
				if (	   ((scaler[scn].z*(img.ix-2*border))<= sx       *scaler[scn].n)
					&& ((scaler[scn].z*(img.iy-2*border))<=(sy0-info)*scaler[scn].n) ) break;
			}
		}

	redisplay:

		sy=sy0-info;

		scn--;
		do
		{
			scn++;
			tx=(scaler[scn].z*(img.ix-2*border))/scaler[scn].n;
			ty=(scaler[scn].z*(img.iy-2*border))/scaler[scn].n;
		}
		while ( !tx || !ty );

		img.dx=tx;
		img.dy=ty;

		if ( tx>=sx ) { x0min=0; x0max=tx-sx; } else x0min=x0max=(tx-sx)/2;
		if ( ty>=sy ) { y0min=0; y0max=ty-sy; } else y0min=y0max=(ty-sy)/2;

		if ( x0>x0max ) x0=x0max;
		if ( y0>y0max ) y0=y0max;
		if ( x0<x0min ) x0=x0min;
		if ( y0<y0min ) y0=y0min;

		vw=tx;	if ( vw>sx ) vw=sx;
		vh=ty;	if ( vh>sy ) vh=sy;
		mx=0;	if ( x0<0  ) mx=-x0;
		my=0;	if ( y0<0  ) my=-y0;

		get_area(scaler+scn,&img,tbuf,buffer,bytepp<4 ? ((uchar *) ximg->data) : NULL,sx,x0<0?0:x0,y0<0?0:y0,vw,vh);

		if ( vh<sy  )
		{
			XClearArea(disp,win,0,0,sx,my,False);
			XClearArea(disp,win,0,my+vh,sx,sy-vh-my,False);
		}

		if ( vw<sx )
		{
			XClearArea(disp,win,0,my,mx,vh,False);
			XClearArea(disp,win,mx+vw,my,sx-vw-mx,vh,False);
		}

		XShmPutImage(disp,win,gcont,ximg,0,0,mx,my,vw,vh,False);

reinfo:

		if ( info )
		{
			i=sprintf(t,"Scale: %i/%i   image %li: %s   size: %ix%i   type: %s %s   %s",
				scaler[scn].z,scaler[scn].n,pn,img.name,img.ix-2*border,img.iy-2*border,img.typ,
				files[pn].flags & flag_tagged ? "  TAGGED" : "        ",
				GetConciseImageInfo(img.name));
			XClearArea(disp,win,0,sy0-info,sx,info,False);
			XDrawImageString(disp, win, gcont, infotx, sy0-infoty, t,i);
		}

		num=0;

		for (;;)
		{
			XNextEvent(disp,&ev);

			switch ( ev.type ) {
				case Expose :
						if ( ev.xexpose.x<mx ) { ev.xexpose.width -=mx-ev.xexpose.x; ev.xexpose.x=mx; }
						if ( ev.xexpose.y<my ) { ev.xexpose.height-=my-ev.xexpose.y; ev.xexpose.y=my; }
						if ( ev.xexpose.x+ev.xexpose.width  > mx+vw ) ev.xexpose.width =mx+vw-ev.xexpose.x;
						if ( ev.xexpose.y+ev.xexpose.height > my+vh ) ev.xexpose.height=my+vh-ev.xexpose.y;
						if ( (ev.xexpose.width>0) && (ev.xexpose.height>0) )
							XShmPutImage(disp,win,gcont,ximg,ev.xexpose.x-mx,ev.xexpose.y-my,ev.xexpose.x,ev.xexpose.y,ev.xexpose.width,ev.xexpose.height,False);
						goto reinfo;

				case ButtonPress :
						bx=ev.xbutton.x; by=ev.xbutton.y;
						break;

				case MotionNotify :
						if ( XEventsQueued(disp,QueuedAfterFlush)>0 ) continue;
				case ButtonRelease :

						tx=ev.xbutton.x-bx;
						ty=ev.xbutton.y-by;
						bx+=tx;
						by+=ty;
						tx*=2; ty*=2;
					tmove:
						x0-=tx;
						y0-=ty;
						if ( x0<x0min ) { tx+=x0-x0min; x0=x0min; }
						if ( y0<y0min ) { ty+=y0-y0min; y0=y0min; }
						if ( x0>x0max ) { tx+=x0-x0max; x0=x0max; }
						if ( y0>y0max ) { ty+=y0-y0max; y0=y0max; }

						if ( !(tx|ty) ) break;

						XCopyArea(disp,win,win,gcont,	tx< 0 ? -tx : mx, ty<0 ? -ty : my,
										tx< 0 ? vw+tx : vw-tx, ty<0 ? vh+ty : vh-ty,
										tx<=0 ? mx : tx, ty<=0 ? my : ty );
						if ( tx>0 )
						{
							get_area(scaler+scn,&img,tbuf,buffer,bytepp<4 ? ((uchar *) ximg->data) : NULL,sx,x0,y0+my,tx,vh);
							XShmPutImage(disp,win,gcont,ximg,0,0,0,my,tx,vh,False);
						}
						if ( ty>0 )
						{
							get_area(scaler+scn,&img,tbuf,buffer,bytepp<4 ? ((uchar *) ximg->data) : NULL,sx,x0+mx,y0,vw,ty);
							XShmPutImage(disp,win,gcont,ximg,0,0,mx,0,vw,ty,False);
						}
						if ( tx<0 )
						{
							get_area(scaler+scn,&img,tbuf,buffer,bytepp<4 ? ((uchar *) ximg->data) : NULL,sx,x0+sx+tx,y0+my,-tx,vh);
							XShmPutImage(disp,win,gcont,ximg,0,0,sx+tx,my,-tx,vh,False);
						}
						if ( ty<0 )
						{
							get_area(scaler+scn,&img,tbuf,buffer,bytepp<4 ? ((uchar *) ximg->data) : NULL,sx,x0+mx,y0+vh+ty,vw,-ty);
							XShmPutImage(disp,win,gcont,ximg,0,0,mx,vh+ty,vw,-ty,False);
						}

						break;

				case KeyPress :
						ks=XLookupKeysym(&(ev.xkey),0);

						if ( (ks>=XK_0) && (ks<=XK_9) && (num<1000000) ) num=10*num+((long) ks)-XK_0; else
						{
							nsc=scn;
							switch ( ks ) {
								case XK_space :
									if ( pn < argc-2 ) { pn+=num?num:1; if ( pn>argc-2 ) pn=argc-2; goto reload; } break;
								case XK_b :
								case XK_BackSpace :
									if ( pn >      0 ) { pn-=num?num:1; if ( pn<     0 ) pn=     0; goto reload; } break;
								case XK_g :
									i = num; if ( i>argc-2 ) i=argc-2;
									if ( pn != i ) { pn=i; goto reload; } break;
								case XK_c :
									tx=x0; ty=y0;
									if ( !mx ) tx=(img.dx-sx)/2;
									if ( !my ) ty=(img.dy-sy)/2;
									if ( (tx==x0) && (ty==y0) ) break;
									if ( (x0-tx<sx) && (tx-x0<sx ) && (y0-ty<sy) && (ty-y0<sy) )
									{
										tx=x0-tx; ty=y0-ty;
										goto tmove;
									}
									else
									{
										x0=tx; y0=ty;
										goto redisplay;
									}
								case XK_h :
									nsc=sc_half;		break;
								case XK_f :
									nsc=sc_full;		break;
								case XK_d :
									nsc=sc_double;		break;
								case XK_j :
									i=num; if ( i>argc-2 ) i=argc-2;
									if ( i!=pn ) { pn=i; goto reload; }
									break;
								case XK_p :
								case XK_plus :
									if ( num ) nsc=scn+num; else nsc=scn+1; if ( nsc>numsc-1 ) nsc=numsc-1;	break;
								case XK_minus :
									if ( num ) nsc=scn-num; else nsc=scn-1; if ( nsc<      0 ) nsc=      0;	break;
								case XK_Home :
									if ( pn!=       0 ) { pn=0;      goto reload;    } break;
								case XK_End :
									if ( pn!=  argc-2 ) { pn=argc-2; goto reload;    } break;
								case XK_l :
									lrot(&img); goto rescale;
								case XK_r :
									rrot(&img); goto rescale;
								case XK_u :
									rot180(&img); goto redisplay;
								case XK_m :
									mirror(&img); goto redisplay;
								case XK_s :
									autoscale=1; goto rescale;
								case XK_t :
									files[pn].flags^=flag_tagged;	goto reinfo;
								case XK_i :
									if ( info )
									{
										info=0;
										if ( y0>y0max-infoheight ) goto redisplay;
										if ( mx ) XClearArea(disp,win,0,sy0-infoheight,sx,infoheight,False);
										scaler[scn].f(tbuf,buffer,4*sx,img.buf,img.ix,x0+mx,y0+sy,vw,infoheight);
										if ( bytepp<4 ) convert16( ((uchar *) ximg->data), buffer, sx, vw, infoheight );
										XShmPutImage(disp,win,gcont,ximg,0,0,mx,sy,vw,infoheight,False);
										sy=vh=sy0;
										y0max-=infoheight;
									}
									else
									{
										info=infoheight;
										if ( my ) goto redisplay;
										i=sprintf(t,"Scale: %i/%i   image %li: %s   size: %ix%i   type: %s %s",
											scaler[scn].z,scaler[scn].n,pn,img.name,img.ix-2*border,img.iy-2*border,img.typ,
											files[pn].flags & flag_tagged ? "  TAGGED" : "");
										XClearArea(disp,win,0,sy0-info,sx,info,False);
										XDrawImageString(disp, win, gcont, infotx, sy0-infoty, t,i);
										sy=vh=sy0-info;
										y0max+=info;
									   }
									break;

								case XK_z :
									XMoveWindow(disp, win, -wzpos[0], -wzpos[1]);
									XRaiseWindow(disp, win);
									goto redisplay;

								case XK_q :
									for (i=0; i<argc-1; i++)
									{
										if ( files[i].flags & flag_deleted ) unlink(files[i].name);
									}
								case XK_Escape :
									for (i=0; i<argc-1; i++)
									{
										if ( files[i].flags & flag_tagged ) printf("%s\n",files[i].name);
									}
									XDestroyImage(ximg);
									shmdt(shminfo.shmaddr);
									shmctl(shminfo.shmid,IPC_RMID,&shmbuf);
									XDestroyWindow(disp,win);
									XCloseDisplay(disp);
									exit(0);
								case XK_Delete :
									files[pn].flags^=flag_deleted;	goto reload;

								case XK_Left  : if ( !(ev.xkey.state & scroll_lock) ) tx= panstep; else tx=-panstep; ty=0; goto tmove;
								case XK_Right : if ( !(ev.xkey.state & scroll_lock) ) tx=-panstep; else tx= panstep; ty=0; goto tmove;
								case XK_Up    : if ( !(ev.xkey.state & scroll_lock) ) ty= panstep; else ty=-panstep; tx=0; goto tmove;
								case XK_Down  : if ( !(ev.xkey.state & scroll_lock) ) ty=-panstep; else ty= panstep; tx=0; goto tmove;

								default : continue;
							}
							num=0;

							if ( nsc!=scn )
							{
								tx=scaler[scn].n*scaler[nsc].z;
								ty=scaler[scn].z*scaler[nsc].n;
								x0=(tx*(2*x0+sx)+ty)/(2*ty) - sx/2;
								y0=(tx*(2*y0+sy)+ty)/(2*ty) - sy/2;
								scn=nsc;
								autoscale=0;
								goto redisplay;
							}
						}

				default : continue;
			}
		}
	}

	return 0;
}


